import itertools

from pytest import fixture, raises, mark
import scipy.sparse as sp
import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal

import quimb as qu

stypes = ("csr", "csc", "bsr", "coo")


@fixture
def od1():
    return qu.rand_matrix(3)


@fixture
def os1():
    return qu.rand_matrix(3, sparse=True, density=0.5)


class TestSparseMatrix:
    @mark.parametrize("stype", stypes)
    def test_simple(self, stype):
        a = qu.core.sparse_matrix([[0, 3], [1, 2]], stype)
        assert a.format == stype
        assert a.dtype == complex


class TestQuimbify:
    def test_vector_create(self):
        x = [1, 2, 3j]
        p = qu.qu(x, qtype='ket')
        assert(type(p) == qu.qarray)
        assert(p.dtype == complex)
        assert(p.shape == (3, 1))
        p = qu.qu(x, qtype='bra')
        assert(p.shape == (1, 3))
        assert_almost_equal(p[0, 2], -3.0j)

    def test_dop_create(self):
        x = np.random.randn(3, 3)
        p = qu.qu(x, qtype='dop')
        assert(type(p) == qu.qarray)
        assert(p.dtype == complex)
        assert(p.shape == (3, 3))

    def test_convert_vector_to_dop(self):
        x = [1, 2, 3j]
        p = qu.qu(x, qtype='r')
        assert_allclose(p, qu.qarray([[1. + 0.j, 2. + 0.j, 0. - 3.j],
                                      [2. + 0.j, 4. + 0.j, 0. - 6.j],
                                      [0. + 3.j, 0. + 6.j, 9. + 0.j]]))

    def test_chopped(self):
        x = [9e-16, 1]
        p = qu.qu(x, 'k', chopped=False)
        assert(p[0, 0] != 0.0)
        p = qu.qu(x, 'k', chopped=True)
        assert(p[0, 0] == 0.0)

    def test_normalized(self):
        x = [3j, 4j]
        p = qu.qu(x, 'k', normalized=False)
        assert_almost_equal(qu.tr(p.H @ p), 25.)
        p = qu.qu(x, 'k', normalized=True)
        assert_almost_equal(qu.tr(p.H @ p), 1.)
        p = qu.qu(x, 'dop', normalized=True)
        assert_almost_equal(qu.tr(p), 1.)

    def test_sparse_create(self):
        x = [[1, 0], [3, 0]]
        p = qu.qu(x, 'dop', sparse=False)
        assert(type(p) == qu.qarray)
        p = qu.qu(x, 'dop', sparse=True)
        assert(type(p) == sp.csr_matrix)
        assert(p.dtype == complex)
        assert(p.nnz == 2)

    def test_sparse_convert_to_dop(self):
        x = [1, 0, 9e-16, 0, 3j]
        p = qu.qu(x, 'ket', sparse=True)
        q = qu.qu(p, 'dop', sparse=True)
        assert(q.shape == (5, 5))
        assert(q.nnz == 9)
        assert_almost_equal(q[4, 4], 9.)
        q = qu.qu(p, 'dop', sparse=True, normalized=True)
        assert_almost_equal(qu.tr(q), 1.)

    @mark.parametrize("qtype, shape, out",
                      (("bra", (1, 4), [[1, 0, 2, -3j]]),
                       ("ket", (4, 1), [[1], [0], [2], [3j]]),
                       ("dop", (4, 4), [[1, 0, 2, -3j],
                                        [0, 0, 0, 0],
                                        [2, 0, 4, -6j],
                                        [3j, 0, 6j, 9]])))
    @mark.parametrize("format_in", stypes)
    @mark.parametrize("format_out", (None,) + stypes)
    @mark.parametrize("dtype", [float, complex, np.float_, np.complex_])
    def test_reshape_sparse(self, qtype, shape, out,
                            format_in, format_out, dtype):
        import warnings

        in_ = [[1], [0], [2], [3j]]
        x = qu.core.sparse_matrix(in_, stype=format_in)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            y = qu.qu(x, qtype=qtype, stype=format_out, dtype=dtype)

        assert y.shape == shape
        assert y.dtype == dtype
        if format_out is None:
            format_out = format_in
        assert y.format == format_out

        if np.issubdtype(dtype, np.floating):
            assert_allclose(y.A, np.real(out), atol=1e-12)
        else:
            assert_allclose(y.A, out)

    @mark.parametrize("qtype, shape, out",
                      (("bra", (1, 4), [[1, 0, 2, -3j]]),
                       ("ket", (4, 1), [[1], [0], [2], [3j]]),
                       ("dop", (4, 4), [[1, 0, 2, -3j],
                                        [0, 0, 0, 0],
                                        [2, 0, 4, -6j],
                                        [3j, 0, 6j, 9]])))
    @mark.parametrize("format_out", (None,) + stypes)
    def test_dense_to_sparse_format(self, qtype, shape, out, format_out):
        x = [[1], [0], [2], [3j]]
        y = qu.qu(x, qtype=qtype, stype=format_out, sparse=True)
        assert y.shape == shape
        assert y.dtype == complex
        if format_out is None:
            format_out = "csr"
        assert y.format == format_out
        assert_allclose(y.A, out)

    @mark.parametrize("qtype, shape",
                      (["bra", (1, 4)],
                       ["ket", (4, 1)],
                       ["dop", (4, 4)]))
    @mark.parametrize("format_out", stypes)
    def test_give_sformat_only(self, qtype, shape, format_out):
        x = [[1], [0], [2], [3j]]
        y = qu.qu(x, qtype=qtype, stype=format_out)
        assert qu.issparse(y)
        assert y.shape == shape
        assert y.format == format_out


class TestInferSize:
    @mark.parametrize("d,base,n",
                      ([8, 2, 3],
                       [16, 2, 4],
                       [9, 3, 2],
                       [81, 3, 4]))
    def test_infer_size(self, d, base, n):
        p = qu.rand_ket(d)
        assert qu.infer_size(p, base) == n

    def test_raises(self):
        p = qu.rand_ket(2) & qu.rand_ket(3)
        with raises(ValueError):
            qu.infer_size(p, base=2)


class TestTrace:
    @mark.parametrize("inpt, outpt",
                      ([[[2, 1], [4, 5]], 7],
                       [[[2, 1], [4, 5j]], 2 + 5j]))
    @mark.parametrize("sparse, func",
                      ([False, qu.core._trace_dense],
                       [True, qu.core._trace_sparse]))
    def test_simple(self, inpt, outpt, sparse, func):
        a = qu.qu(inpt, sparse=sparse)
        assert(qu.trace(a) == outpt)


class TestITrace:
    @mark.parametrize("axes", [(0, 1), ((0,), (1,))])
    def test_axes_types(self, axes):
        a = qu.rand_matrix(4)
        b = qu.itrace(a, axes)
        assert_allclose(b, np.trace(a))

    def test_complex_dims(self):
        a = np.random.rand(4, 3, 2, 2, 4, 3)
        atr = qu.itrace(a, ((0, 1, 2), (4, 5, 3)))
        btr = np.trace(np.trace(np.trace(a, axis1=1, axis2=5),
                                axis1=1, axis2=2))
        assert_allclose(atr, btr)


class TestNormalize:
    def test_normalize_ket(self):
        a = qu.qu([1, -1j], 'ket')
        b = qu.nmlz(a, inplace=False)
        assert_almost_equal(qu.trace(b.H @ b), 1.0)
        assert_almost_equal(qu.trace(a.H @ a), 2.0)

    def test_normalize_bra(self):
        a = qu.qu([1, -1j], 'bra')
        b = qu.nmlz(a, inplace=False)
        assert_almost_equal(qu.trace(b @ b.H), 1.0)

    def test_normalize_dop(self):
        a = qu.qu([1, -1j], 'dop')
        b = qu.nmlz(a, inplace=False)
        assert_almost_equal(qu.trace(b), 1.0)

    def test_normalize_inplace_ket(self):
        a = qu.qu([1, -1j], 'ket')
        a.nmlz(inplace=True)
        assert_almost_equal(qu.trace(a.H @ a), 1.0)

    def test_normalize_inplace_bra(self):
        a = qu.qu([1, -1j], 'bra')
        a.nmlz(inplace=True)
        assert_almost_equal(qu.trace(a @ a.H), 1.0)

    def test_normalize_inplace_dop(self):
        a = qu.qu([1, -1j], 'dop')
        b = qu.nmlz(a, inplace=True)
        assert_almost_equal(qu.trace(a), 1.0)
        assert_almost_equal(qu.trace(b), 1.0)


class TestDimMap:
    @mark.parametrize("numpy", [False, True])
    def test_1d(self, numpy):
        dims = [10, 11, 12, 13]
        coos = (1, 2, 3)
        if numpy:
            dims, coos = np.asarray(dims), np.asarray(coos)
        ndims, ncoos = qu.dim_map(dims, coos)
        assert_allclose([ndims[c] for c in ncoos], (11, 12, 13))
        coos = ([-1], [2], [5])
        with raises(ValueError):
            ndims, ncoos = qu.dim_map(dims, coos)
        ndims, ncoos = qu.dim_map(dims, coos, cyclic=True)
        assert_allclose([ndims[c] for c in ncoos], (13, 12, 11))
        ndims, ncoos = qu.dim_map(dims, coos, trim=True)
        assert_allclose([ndims[c] for c in ncoos], [12])

    def test_2d(self):
        dims = [[200, 201, 202, 203],
                [210, 211, 212, 213]]
        coos = ((1, 2), (1, 3), (0, 3))
        ndims, ncoos = qu.dim_map(dims, coos)
        assert_allclose([ndims[c] for c in ncoos], (212, 213, 203))
        coos = ((-1, 1), (1, 2), (3, 4))
        with raises(ValueError):
            ndims, ncoos = qu.dim_map(dims, coos)
        ndims, ncoos = qu.dim_map(dims, coos, cyclic=True)
        assert_allclose([ndims[c] for c in ncoos], (211, 212, 210))
        ndims, ncoos = qu.dim_map(dims, coos, trim=True)
        assert_allclose([ndims[c] for c in ncoos], [212])

    def test_3d(self):
        dims = [[[3000, 3001, 3002],
                 [3010, 3011, 3012],
                 [3020, 3021, 3022]],
                [[3100, 3101, 3102],
                 [3110, 3111, 3112],
                 [3120, 3121, 3122]]]
        coos = ((0, 0, 2), (1, 1, 2), (1, 2, 0))
        ndims, ncoos = qu.dim_map(dims, coos)
        assert_allclose([ndims[c] for c in ncoos], (3002, 3112, 3120))
        coos = ((0, -1, 2), (1, 2, 2), (4, -1, 3))
        with raises(ValueError):
            ndims, ncoos = qu.dim_map(dims, coos)
        ndims, ncoos = qu.dim_map(dims, coos, cyclic=True)
        assert_allclose([ndims[c] for c in ncoos], (3022, 3122, 3020))
        ndims, ncoos = qu.dim_map(dims, coos, trim=True)
        assert_allclose([ndims[c] for c in ncoos], [3122])


class TestDimCompress:
    def test_edge(self):
        dims = [2, 3, 2, 4, 5]
        coos = [0, 4]
        ndims, ncoos = qu.dim_compress(dims, coos)
        assert ndims == (2, 24, 5)
        assert ncoos == (0, 2)

    def test_middle(self):
        dims = [5, 3, 2, 5, 4, 3, 2]
        coos = [1, 2, 3, 5]
        ndims, ncoos = qu.dim_compress(dims, coos)
        assert ndims == (5, 30, 4, 3, 2)
        assert ncoos == (1, 3)

    def test_single(self):
        dims = [5, 3, 2, 5, 4, 3, 2]
        coos = 3
        ndims, ncoos = qu.dim_compress(dims, coos)
        assert ndims == (30, 5, 24)
        assert ncoos == (1,)

    @mark.parametrize("dims, inds, ndims, ninds",
                      [([2, 2], [0, 1], (4,), (0,)),
                       ([4], [0], (4,), (0,))])
    def test_tiny(self, dims, inds, ndims, ninds):
        dims, inds = qu.dim_compress(dims, inds)
        assert dims == ndims
        assert inds == ninds


class TestEye:
    def test_eye_dense(self):
        a = qu.eye(3, sparse=False)
        assert a.shape == (3, 3)
        assert isinstance(a, qu.qarray)
        assert a.dtype == complex

    def test_eye_sparse(self):
        a = qu.eye(3, sparse=True)
        assert a.shape == (3, 3)
        assert isinstance(a, sp.csr_matrix)
        assert a.dtype == complex


class TestKron:
    @mark.parametrize("parallel", [True, False])
    def test_kron_basic(self, parallel):
        a = qu.rand_ket(2)
        b = qu.rand_ket(4)
        c = qu.rand_ket(4)
        d = qu.rand_ket(5)
        t = qu.kron(a, b, c, d, parallel=parallel)
        assert_allclose(t, a & b & c & d)

    @mark.parametrize("sparse", [True, False])
    @mark.parametrize("ri,rf", ([0, 4], [75, 89], [150, 168], [0, 168]))
    def test_kron_ownership(self, sparse, ri, rf):
        dims = [7, 2, 4, 3]
        ops = [qu.rand_matrix(d, sparse=sparse) for d in dims]
        X1 = qu.kron(*ops)[ri:rf, :]
        X2 = qu.kron(*ops, ownership=(ri, rf))
        assert_allclose(X1.A, X2.A)


class Testikron:
    def test_basic(self):
        a = qu.rand_matrix(2)
        i = qu.eye(2)
        dims = [2, 2, 2]
        b = qu.ikron([a], dims, [0])
        assert_allclose(b, a & i & i)
        b = qu.ikron([a], dims, [1])
        assert_allclose(b, i & a & i)
        b = qu.ikron([a], dims, [2])
        assert_allclose(b, i & i & a)
        b = qu.ikron([a], dims, [0, 2])
        assert_allclose(b, a & i & a)
        b = qu.ikron([a], dims, [0, 1, 2])
        assert_allclose(b, a & a & a)

    def test_mid_multi(self):
        a = [qu.rand_matrix(2) for i in range(3)]
        i = qu.eye(2)
        dims = [2, 2, 2, 2, 2, 2]
        inds = [1, 2, 4]
        b = qu.ikron(a, dims, inds)
        assert_allclose(b, i & a[0] & a[1] & i & a[2] & i)

    def test_mid_multi_reverse(self):
        a = [qu.rand_matrix(2) for i in range(3)]
        i = qu.eye(2)
        dims = [2, 2, 2, 2, 2, 2]
        inds = [5, 4, 1]
        b = qu.ikron(a, dims, inds)
        assert_allclose(b, i & a[2] & i & i & a[1] & a[0])

    def test_auto(self):
        a = qu.rand_matrix(2)
        i = qu.eye(2)
        b = qu.ikron([a], (2, -1, 2), [1])
        assert_allclose(b, i & a & i)

    def test_ndarrays(self):
        a = qu.rand_matrix(2)
        i = qu.eye(2)
        b = qu.ikron([a], np.array([2, 2, 2]), [0, 2])
        assert_allclose(b, a & i & a)
        b = qu.ikron([a], [2, 2, 2], np.array([0, 2]))
        assert_allclose(b, a & i & a)

    def test_overlap(self):
        a = [qu.rand_matrix(4) for i in range(2)]
        dims1 = [2, 2, 2, 2, 2, 2]
        dims2 = [2, 4, 4, 2]
        b = qu.ikron(a, dims1, [1, 2, 3, 4])
        c = qu.ikron(a, dims2, [1, 2])
        assert_allclose(c, b)
        dims2 = [4, 2, 2, 4]
        b = qu.ikron(a, dims1, [0, 1, 4, 5])
        c = qu.ikron(a, dims2, [0, 3])
        assert_allclose(c, b)

    def test_holey_overlap(self):
        a = qu.rand_matrix(8)
        dims1 = (2, 2, 2, 2, 2)
        dims2 = (2, 8, 2)
        b = qu.ikron(a, dims1, (1, 3))
        c = qu.ikron(a, dims2, 1)
        assert_allclose(b, c)
        dims1 = (2, 2, 2, 2, 2)
        dims2 = (2, 2, 8)
        b = qu.ikron(a, dims1, (2, 4))
        c = qu.ikron(a, dims2, 2)
        assert_allclose(b, c)
        dims1 = (2, 2, 2, 2, 2)
        dims2 = (8, 2, 2)
        b = qu.ikron(a, dims1, (0, 2))
        c = qu.ikron(a, dims2, 0)
        assert_allclose(b, c)

    def test_sparse(self):
        i = qu.eye(2, sparse=True)
        a = qu.qu(qu.rand_matrix(2), sparse=True)
        b = qu.ikron(a, [2, 2, 2], 1)  # infer sparse
        assert(qu.issparse(b))
        assert_allclose(b.A, (i & a & i).A)
        a = qu.rand_matrix(2)
        b = qu.ikron(a, [2, 2, 2], 1, sparse=True)  # explicit sparse
        assert(qu.issparse(b))
        assert_allclose(b.A, (i & a & i).A)

    def test_2d_simple(self):
        a = (qu.rand_matrix(2), qu.rand_matrix(2))
        dims = ((2, 3), (3, 2))
        inds = ((0, 0), (1, 1))
        b = qu.ikron(a, dims, inds)
        assert b.shape == (36, 36)
        assert_allclose(b, a[0] & qu.eye(9) & a[1])

    @mark.parametrize("stype", (None,) + stypes)
    @mark.parametrize("pos", [0, 1, 2,
                              (0,), (1,), (2,),
                              (0, 1),
                              (1, 2),
                              (0, 2)])
    @mark.parametrize("coo_build", [False, True])
    def test_sparse_format_outputs(self, os1, stype, pos, coo_build):
        x = qu.ikron(os1, [3, 3, 3], pos,
                     stype=stype, coo_build=coo_build)
        assert x.format == "csr" if stype is None else stype

    @mark.parametrize("stype", (None,) + stypes)
    @mark.parametrize("pos", [0, 1, 2,
                              (0,), (1,), (2,),
                              (0, 1),
                              (1, 2),
                              (0, 2)])
    @mark.parametrize("coo_build", [False, True])
    def test_sparse_format_outputs_with_dense(self, od1, stype, pos,
                                              coo_build):
        x = qu.ikron(od1, [3, 3, 3], pos, sparse=True,
                     stype=stype, coo_build=coo_build)
        try:
            default = "bsr" if (2 in pos and not coo_build) else "csr"
        except TypeError:
            default = "bsr" if (pos == 2 and not coo_build) else "csr"
        assert x.format == default if stype is None else stype

    @mark.parametrize("sparse", [True, False])
    @mark.parametrize("ri,rf", ([0, 4], [75, 89], [150, 168], [0, 168]))
    def test_ikron_ownership(self, sparse, ri, rf):
        dims = [7, 2, 4, 3]
        X = qu.rand_matrix(2, sparse=sparse)
        Y = qu.rand_matrix(3, sparse=sparse)
        X1 = qu.ikron((X, Y), dims, (1, 3))[ri:rf, :]
        X2 = qu.ikron((X, Y), dims, (1, 3), ownership=(ri, rf))
        assert_allclose(X1.A, X2.A)


class TestPermikron:
    def test_dop_spread(self):
        a = qu.rand_rho(4)
        b = qu.pkron(a, [2, 2, 2], [0, 2])
        c = ((a & qu.eye(2)).A
             .reshape([2, 2, 2, 2, 2, 2])
             .transpose([0, 2, 1, 3, 5, 4])
             .reshape([8, 8]))
        assert_allclose(b, c)

    def test_dop_reverse(self):
        a = qu.rand_rho(4)
        b = qu.pkron(a, np.array([2, 2, 2]), [2, 0])
        c = ((a & qu.eye(2)).A.reshape([2, 2, 2, 2, 2, 2])
                            .transpose([1, 2, 0, 4, 5, 3])
                            .reshape([8, 8]))
        assert_allclose(b, c)

    def test_dop_reverse_sparse(self):
        a = qu.rand_rho(4, sparse=True, density=0.5)
        b = qu.pkron(a, np.array([2, 2, 2]), [2, 0])
        c = ((a & qu.eye(2)).A.reshape([2, 2, 2, 2, 2, 2])
                            .transpose([1, 2, 0, 4, 5, 3])
                            .reshape([8, 8]))
        assert_allclose(b.A, c)


class TestPermute:
    def test_permute_ket(self):
        a = qu.up() & qu.plus() & qu.yplus()
        b = qu.permute(a, [2, 2, 2], [2, 0, 1])
        assert_allclose(b, qu.yplus() & qu.up() & qu.plus())

    def test_permute_op(self):
        a = qu.pauli('x') & qu.pauli('y') & qu.pauli('z')
        b = qu.permute(a, [2, 2, 2], [2, 0, 1])
        assert_allclose(b, qu.pauli('z') & qu.pauli('x') & qu.pauli('y'))

    def test_entangled_permute(self):
        dims = [2, 2, 2]
        a = qu.bell_state(0) & qu.up()
        assert_allclose(qu.mutinf_subsys(a, dims, 0, 1), 2.)
        b = qu.permute(a, dims, [1, 2, 0])
        assert_allclose(qu.mutinf_subsys(b, dims, 0, 1), 0., atol=1e-12)
        assert_allclose(qu.mutinf_subsys(b, dims, 0, 2), 2.)

    def test_permute_sparse_ket(self):
        dims = [3, 2, 5, 4]
        a = qu.rand_ket(qu.prod(dims), sparse=True, density=0.5)
        b = qu.permute(a, dims, [3, 1, 2, 0])
        c = qu.permute(a.A, dims, [3, 1, 2, 0])
        assert_allclose(b.A, c)

    def test_permute_sparse_op(self):
        dims = [3, 2, 5, 4]
        a = qu.rand_rho(qu.prod(dims), sparse=True, density=0.5)
        b = qu.permute(a, dims, [3, 1, 2, 0])
        c = qu.permute(a.A, dims, [3, 1, 2, 0])
        assert_allclose(b.A, c)


class TestPartialTraceDense:
    def test_partial_trace_basic(self):
        a = qu.rand_rho(2**2)
        b = qu.partial_trace(a, [2, 2], 0)
        assert isinstance(b, qu.qarray)
        assert qu.isherm(b)
        assert_allclose(qu.tr(b), 1.0)

    def test_ptr_compare_to_manual(self):
        a = qu.rand_rho(2**2)
        b = qu.partial_trace(a, [2, 2], 0)
        c = a.A.reshape([2, 2, 2, 2]).trace(axis1=1, axis2=3)
        assert_allclose(b, c)
        b = qu.partial_trace(a, [2, 2], 1)
        c = a.A.reshape([2, 2, 2, 2]).trace(axis1=0, axis2=2)
        assert_allclose(b, c)

    def test_partial_trace_early_return(self):
        a = qu.qu([0.5, 0.5, 0.5, 0.5], 'ket')
        b = qu.partial_trace(a, [2, 2], [0, 1])
        assert_allclose(a @ a.H, b)
        a = qu.qu([0.5, 0.5, 0.5, 0.5], 'dop')
        b = qu.partial_trace(a, [2, 2], [0, 1])
        assert_allclose(a, b)

    def test_partial_trace_return_type(self):
        a = qu.qu([0, 2**-0.5, 2**-0.5, 0], 'ket')
        b = qu.partial_trace(a, [2, 2], 1)
        assert(type(b) == qu.qarray)
        a = qu.qu([0, 2**-0.5, 2**-0.5, 0], 'dop')
        b = qu.partial_trace(a, [2, 2], 1)
        assert(type(b) == qu.qarray)

    def test_partial_trace_single_ket(self):
        dims = [2, 3, 4]
        a = np.random.randn(qu.prod(dims), 1)
        for i, dim in enumerate(dims):
            b = qu.partial_trace(a, dims, i)
            assert(b.shape[0] == dim)

    def test_partial_trace_multi_ket(self):
        dims = [2, 3, 4]
        a = np.random.randn(qu.prod(dims), 1)
        for i1, i2 in itertools.combinations([0, 1, 2], 2):
            b = qu.partial_trace(a, dims, [i1, i2])
            assert(b.shape[1] == dims[i1] * dims[i2])

    def test_partial_trace_dop_product_state(self):
        dims = [3, 2, 4, 2, 3]
        ps = [qu.rand_rho(dim) for dim in dims]
        pt = qu.kron(*ps)
        for i, dim in enumerate(dims):
            p = qu.partial_trace(pt, dims, i)
            assert_allclose(p, ps[i])

    def test_partial_trace_bell_states(self):
        for lab in ('psi-', 'psi+', 'phi-', 'phi+'):
            psi = qu.bell_state(lab, qtype='dop')
            rhoa = qu.partial_trace(psi, [2, 2], 0)
            assert_allclose(rhoa, qu.eye(2) / 2)

    def test_partial_trace_supply_ndarray(self):
        a = qu.rand_rho(2**3)
        dims = np.array([2, 2, 2])
        keep = np.array(1)
        b = qu.partial_trace(a, dims, keep)
        assert(b.shape[0] == 2)

    def test_partial_trace_order_doesnt_matter(self):
        a = qu.rand_rho(2**3)
        dims = np.array([2, 2, 2])
        b1 = qu.partial_trace(a, dims, [0, 2])
        b2 = qu.partial_trace(a, dims, [2, 0])
        assert_allclose(b1, b2)


class TestTraceLose:
    def test_rps(self):
        a, b, c = (qu.rand_rho(2, sparse=True, density=0.5),
                   qu.rand_rho(3, sparse=True, density=0.5),
                   qu.rand_rho(2, sparse=True, density=0.5))
        abc = a & b & c
        pab = qu.core._trace_lose(abc, [2, 3, 2], 2)
        assert_allclose(pab, (a & b).A)
        pac = qu.core._trace_lose(abc, [2, 3, 2], 1)
        assert_allclose(pac, (a & c).A)
        pbc = qu.core._trace_lose(abc, [2, 3, 2], 0)
        assert_allclose(pbc, (b & c).A)

    def test_bell_state(self):
        a = qu.bell_state('psi-', sparse=True)
        b = qu.core._trace_lose(a @ a.H, [2, 2], 0)
        assert_allclose(b, qu.eye(2) / 2)
        b = qu.core._trace_lose(a @ a.H, [2, 2], 1)
        assert_allclose(b, qu.eye(2) / 2)

    def test_vs_ptr(self):
        a = qu.rand_rho(6, sparse=True, density=0.5)
        b = qu.core._trace_lose(a, [2, 3], 1)
        c = qu.partial_trace(a.A, [2, 3], 0)
        assert_allclose(b, c)
        b = qu.core._trace_lose(a, [2, 3], 0)
        c = qu.partial_trace(a.A, [2, 3], 1)
        assert_allclose(b, c)

    def test_vec_dense(self):
        a = qu.rand_ket(4)
        b = qu.core._trace_lose(a, [2, 2], 1)
        c = qu.partial_trace(a.A, [2, 2], 0)
        assert_allclose(b, c)
        b = qu.core._trace_lose(a, [2, 2], 0)
        c = qu.partial_trace(a.A, [2, 2], 1)
        assert_allclose(b, c)


class TestTraceKeep:
    def test_rps(self):
        a, b, c = (qu.rand_rho(2, sparse=True, density=0.5),
                   qu.rand_rho(3, sparse=True, density=0.5),
                   qu.rand_rho(2, sparse=True, density=0.5))
        abc = a & b & c
        pc = qu.core._trace_keep(abc, [2, 3, 2], 2)
        assert_allclose(pc, c.A)
        pb = qu.core._trace_keep(abc, [2, 3, 2], 1)
        assert_allclose(pb, b.A)
        pa = qu.core._trace_keep(abc, [2, 3, 2], 0)
        assert_allclose(pa, a.A)

    def test_bell_state(self):
        a = qu.bell_state('psi-', sparse=True)
        b = qu.core._trace_keep(a @ a.H, [2, 2], 0)
        assert_allclose(b, qu.eye(2) / 2)
        b = qu.core._trace_keep(a @ a.H, [2, 2], 1)
        assert_allclose(b, qu.eye(2) / 2)

    def test_vs_ptr(self):
        a = qu.rand_rho(6, sparse=True, density=0.5)
        b = qu.core._trace_keep(a, [2, 3], 0)
        c = qu.partial_trace(a.A, [2, 3], 0)
        assert_allclose(b, c)
        b = qu.core._trace_keep(a, [2, 3], 1)
        c = qu.partial_trace(a.A, [2, 3], 1)
        assert_allclose(b, c)

    def test_vec_dense(self):
        a = qu.rand_ket(4)
        b = qu.core._trace_keep(a, [2, 2], 0)
        c = qu.partial_trace(a.A, [2, 2], 0)
        assert_allclose(b, c)
        b = qu.core._trace_keep(a, [2, 2], 1)
        c = qu.partial_trace(a.A, [2, 2], 1)
        assert_allclose(b, c)


class TestPartialTraceSparse:
    def test_partial_trace_sparse_basic(self):
        a = qu.rand_rho(4)
        b = qu.partial_trace(a, [2, 2], 0)
        assert type(b) == qu.qarray
        assert qu.isherm(b)
        assert_allclose(qu.tr(b), 1.0)

    def test_partial_trace_simple_single(self):
        a = qu.rand_rho(12, sparse=True, density=0.5)
        dims = [2, 3, 2]
        b = qu.partial_trace(a, dims, 1)
        c = a.A.reshape([*dims, *dims])  \
             .trace(axis1=2, axis2=5)  \
             .trace(axis1=0, axis2=2)
        assert_allclose(c, b)

    def test_partial_trace_simple_double(self):
        a = qu.rand_rho(12, sparse=True, density=0.5)
        dims = [2, 3, 2]
        b = qu.partial_trace(a, dims, [0, 2])
        c = qu.partial_trace(a.A, dims, [0, 2])
        assert_allclose(b, c)
        b = qu.partial_trace(a, dims, [1, 2])
        c = qu.partial_trace(a.A, dims, [1, 2])
        assert_allclose(b, c)

    def test_partial_trace_simple_ket(self):
        a = qu.rand_ket(12, sparse=True, density=0.5)
        dims = [2, 3, 2]
        b = qu.partial_trace(a, dims, [0, 1])
        c = qu.partial_trace(a.A, dims, [0, 1])
        assert_allclose(b, c)


class TestChop:
    def test_chop_inplace(self):
        a = qu.qu([-1j, 0.1 + 0.2j])
        qu.chop(a, tol=0.11, inplace=True)
        assert_allclose(a, qu.qu([-1j, 0.2j]))
        # Sparse
        a = qu.qu([-1j, 0.1 + 0.2j], sparse=True)
        qu.chop(a, tol=0.11, inplace=True)
        b = qu.qu([-1j, 0.2j], sparse=True)
        assert((a != b).nnz == 0)

    def test_chop_inplace_dop(self):
        a = qu.qu([1, 0.1], 'dop')
        qu.chop(a, tol=0.11, inplace=True)
        assert_allclose(a, qu.qu([1, 0], 'dop'))
        a = qu.qu([1, 0.1], 'dop', sparse=True)
        qu.chop(a, tol=0.11, inplace=True)
        b = qu.qu([1, 0.0], 'dop', sparse=True)
        assert((a != b).nnz == 0)

    def test_chop_copy(self):
        a = qu.qu([-1j, 0.1 + 0.2j])
        b = qu.chop(a, tol=0.11, inplace=False)
        assert_allclose(a, qu.qu([-1j, 0.1 + 0.2j]))
        assert_allclose(b, qu.qu([-1j, 0.2j]))
        # Sparse
        a = qu.qu([-1j, 0.1 + 0.2j], sparse=True)
        b = qu.chop(a, tol=0.11, inplace=False)
        ao = qu.qu([-1j, 0.1 + 0.2j], sparse=True)
        bo = qu.qu([-1j, 0.2j], sparse=True)
        assert((a != ao).nnz == 0)
        assert((b != bo).nnz == 0)


class TestExpec:
    @mark.parametrize("qtype1", ['ket', 'dop'])
    @mark.parametrize("spars1", [True, False])
    @mark.parametrize("qtype2", ['ket', 'dop'])
    @mark.parametrize("spars2", [True, False])
    def test_all(self, qtype1, spars1, qtype2, spars2):
        a = qu.qu([[1], [2j], [3]], qtype=qtype1, sparse=spars1)
        b = qu.qu([[1j], [2], [3j]], qtype=qtype2, sparse=spars2)
        c = qu.expec(a, b)
        assert not isinstance(c, complex)
        assert_allclose(c, 36)

    @mark.parametrize("qtype", ['ket', 'dop'])
    @mark.parametrize("sparse", [True, False])
    @mark.parametrize("s", ['x', 'y', 'z'])
    def test_negative_expec(self, qtype, sparse, s):
        a = qu.singlet(qtype=qtype)
        b = qu.pauli(s, sparse=sparse) & qu.pauli(s, sparse=sparse)
        assert_allclose(qu.expec(a, b), -1)


class TestNumbaFuncs:

    @mark.parametrize("size", [300, 3000, (300, 5), (3000, 5)])
    @mark.parametrize("X_dtype", ['float32', 'float64',
                                  'complex64', 'complex128'])
    @mark.parametrize("c_dtype", ['float32', 'float64'])
    def test_subtract_update(
        self, size, X_dtype, c_dtype,
    ):
        X = qu.randn(size, dtype=X_dtype)
        Y = qu.randn(size, dtype=X_dtype)
        c = qu.randn(1, dtype=c_dtype).item()
        res = X - c * Y
        qu.core.subtract_update_(X, c, Y)
        assert_allclose(res, X)

    @mark.parametrize("size", [300, 3000, (300, 5), (3000, 5)])
    @mark.parametrize("X_dtype", ['float32', 'float64',
                                  'complex64', 'complex128'])
    @mark.parametrize("c_dtype", ['float32', 'float64'])
    def test_divide_update(
        self, size, X_dtype, c_dtype,
    ):
        X = qu.randn(size, dtype=X_dtype)
        Y = np.empty_like(X)
        c = qu.randn(1, dtype=c_dtype).item()
        res = X / c
        qu.core.divide_update_(X, c, Y)
        assert_allclose(res, Y, rtol=1e-6)



from math import log2
import pytest
import numpy as np
from numpy.testing import assert_allclose

from quimb import (
    seed_rand,
    approx_spectral_function,
    eigvalsh,
    logneg_subsys,
)
from quimb.tensor import MPO_rand_herm, MPO_ham_heis, DMRG2
from quimb.tensor.tensor_approx_spectral import construct_lanczos_tridiag_MPO


np.random.seed(42)


# XXX: these all need improvement


class TestMPOSpectralApprox:

    def test_constructing_tridiag_works(self):
        A = MPO_rand_herm(10, 7)
        for _ in construct_lanczos_tridiag_MPO(A, 5):
            pass

    @pytest.mark.parametrize("fn", [abs, np.cos, lambda x: np.sin(x)**2])
    def test_approx_fn(self, fn):
        A = MPO_rand_herm(10, 7, normalize=True)
        xe = sum(fn(eigvalsh(A.to_dense())))
        xf = approx_spectral_function(A, fn, tol=0.1, verbosity=2)
        assert_allclose(xe, xf, rtol=0.5)

    def test_realistic(self):
        seed_rand(42)
        ham = MPO_ham_heis(20)
        dmrg = DMRG2(ham, bond_dims=[2, 4])
        dmrg.solve()
        rho_ab = dmrg.state.ptr(range(6, 14))
        xf = approx_spectral_function(rho_ab, lambda x: x,
                                      tol=0.1, verbosity=2)
        assert_allclose(1.0, xf, rtol=0.6, atol=0.001)

    def test_realistic_ent(self):
        n = 12
        sysa, sysb = range(3, 6), range(6, 8)
        sysab = (*sysa, *sysb)

        ham = MPO_ham_heis(n)
        dmrg = DMRG2(ham, bond_dims=[10])
        dmrg.solve()

        psi0 = dmrg.state.to_dense()
        lne = logneg_subsys(psi0, [2] * n, sysa=sysa, sysb=sysb)

        rho_ab = dmrg.state.ptr(sysab, rescale_sites=True)
        rho_ab_pt = rho_ab.partial_transpose(range(3))
        lnx = log2(approx_spectral_function(rho_ab_pt, abs,
                                            tol=0.1, verbosity=2))
        assert_allclose(lne, lnx, rtol=0.6, atol=0.1)



"""Tools for interfacing the tensor and tensor network objects with other
libraries.
"""

import functools

from ..utils import tree_map
from .tensor_core import Tensor, TensorNetwork


class Placeholder:

    __slots__ = ("shape",)

    def __init__(self, x):
        self.shape = getattr(x, "shape", None)

    def __repr__(self):
        return f"Placeholder(shape={self.shape})"


def pack(obj):
    """Take a tensor or tensor network like object and return a skeleton needed
    to reconstruct it, and a pytree of raw parameters.

    Parameters
    ----------
    obj : Tensor, TensorNetwork, or similar
        Something that has ``copy``, ``set_params``, and ``get_params``
        methods.

    Returns
    -------
    params : pytree
        A pytree of raw parameter arrays.
    skeleton : Tensor, TensorNetwork, or similar
        A copy of ``obj`` with all references to the original data removed.
    """
    skeleton = obj.copy()
    params = skeleton.get_params()
    placeholders = tree_map(Placeholder, params)
    skeleton.set_params(placeholders)
    return params, skeleton


def unpack(params, skeleton):
    """Take a skeleton of a tensor or tensor network like object and a pytree
    of raw parameters and return a new reconstructed object with those
    parameters inserted.

    Parameters
    ----------
    params : pytree
        A pytree of raw parameter arrays, with the same structure as the
        output of ``skeleton.get_params()``.
    skeleton : Tensor, TensorNetwork, or similar
        Something that has ``copy``, ``set_params``, and ``get_params``
        methods.

    Returns
    -------
    obj : Tensor, TensorNetwork, or similar
        A copy of ``skeleton`` with parameters inserted.
    """
    obj = skeleton.copy()
    obj.set_params(params)
    return obj


# -------------------------------- jax -------------------------------------- #


_JAX_REGISTERED_TN_CLASSES = set()


def jax_pack(obj):
    # jax requires the top level children to be a tuple
    params, aux = pack(obj)
    children = (params,)
    return children, aux


def jax_unpack(aux, children):
    # jax also flips the return order from above
    (params,) = children
    return unpack(params, aux)


def jax_register_pytree():
    import jax

    queue = [Tensor, TensorNetwork]
    while queue:
        cls = queue.pop()
        if cls not in _JAX_REGISTERED_TN_CLASSES:
            jax.tree_util.register_pytree_node(cls, jax_pack, jax_unpack)
            _JAX_REGISTERED_TN_CLASSES.add(cls)
        queue.extend(cls.__subclasses__())


@functools.lru_cache(1)
def get_jax():
    import jax

    jax_register_pytree()
    return jax



import numpy as np

from ..utils import ensure_dict, continuous_progbar, deprecated
from ..utils import progbar as Progbar
from .array_ops import norm_fro
from .tensor_arbgeom_tebd import LocalHamGen


class LocalHam1D(LocalHamGen):
    """An simple interacting hamiltonian object used, for instance, in TEBD.
    Once instantiated, the ``LocalHam1D`` hamiltonian stores a single term
    per pair of sites, cached versions of which can be retrieved like
    ``H.get_gate_expm((i, i + 1), -1j * 0.5)`` etc.

    Parameters
    ----------
    L : int
        The size of the hamiltonian.
    H2 : array_like or dict[tuple[int], array_like]
        The sum of interaction terms. If a dict is given, the keys should be
        nearest neighbours like ``(10, 11)``, apart from any default term which
        should have the key ``None``, and the values should be the sum of
        interaction terms for that interaction.
    H1 : array_like or dict[int, array_like], optional
        The sum of single site terms. If a dict is given, the keys should be
        integer sites, apart from any default term which should have the key
        ``None``, and the values should be the sum of single site terms for
        that site.
    cyclic : bool, optional
        Whether the hamiltonian has periodic boundary conditions or not.

    Attributes
    ----------
    terms : dict[tuple[int], array]
        The terms in the hamiltonian, combined from the inputs such that there
        is a single term per pair.

    Examples
    --------
    A simple, translationally invariant, interaction-only ``LocalHam1D``::

        >>> XX = pauli('X') & pauli('X')
        >>> YY = pauli('Y') & pauli('Y')
        >>> ham = LocalHam1D(L=100, H2=XX + YY)

    The same, but with a translationally invariant field as well::

        >>> Z = pauli('Z')
        >>> ham = LocalHam1D(L=100, H2=XX + YY, H1=Z)

    Specifying a default interaction and field, with custom values set for some
    sites::

        >>> H2 = {None: XX + YY, (49, 50): (XX + YY) / 2}
        >>> H1 = {None: Z, 49: 2 * Z, 50: 2 * Z}
        >>> ham = LocalHam1D(L=100, H2=H2, H1=H1)

    Specifying the hamiltonian entirely through site specific interactions and
    fields::

        >>> H2 = {(i, i + 1): XX + YY for i in range(99)}
        >>> H1 = {i: Z for i in range(100)}
        >>> ham = LocalHam1D(L=100, H2=H2, H1=H1)

    See Also
    --------
    SpinHam1D
    """

    def __init__(self, L, H2, H1=None, cyclic=False):
        self.L = int(L)
        self.cyclic = cyclic

        # parse two site terms
        if hasattr(H2, 'shape'):
            # use as default nearest neighbour term
            H2 = {None: H2}
        else:
            H2 = dict(H2)

        default_H2 = H2.pop(None, None)
        if default_H2 is not None:
            for i in range(self.L + int(self.cyclic) - 1):
                coo_a = i
                coo_b = (i + 1) % self.L
                if (coo_a, coo_b) not in H2 and (coo_b, coo_a) not in H2:
                    H2[coo_a, coo_b] = default_H2

        super().__init__(H2=H2, H1=H1)

    def mean_norm(self):
        """Computes the average frobenius norm of local terms.
        """
        return sum(
            norm_fro(h)
            for h in self.terms.values()
        ) / len(self.terms)

    def __repr__(self):
        return f"<LocalHam1D(L={self.L}, cyclic={self.cyclic})>"


NNI = deprecated(LocalHam1D, 'NNI', 'LocalHam1D')


class TEBD:
    """Class implementing Time Evolving Block Decimation (TEBD) [1].

    [1] Guifré Vidal, Efficient Classical Simulation of Slightly Entangled
    Quantum Computations, PRL 91, 147902 (2003)

    Parameters
    ----------
    p0 : MatrixProductState
        Initial state.
    H : LocalHam1D or array_like
        Dense hamiltonian representing the two body interaction. Should have
        shape ``(d * d, d * d)``, where ``d`` is the physical dimension of
        ``p0``.
    dt : float, optional
        Default time step, cannot be set as well as ``tol``.
    tol : float, optional
        Default target error for each evolution, cannot be set as well as
        ``dt``, which will instead be calculated from the trotter orderm length
        of time, and hamiltonian norm.
    t0 : float, optional
        Initial time. Defaults to 0.0.
    split_opts : dict, optional
        Compression options applied for splitting after gate application, see
        :func:`~quimb.tensor.tensor_core.tensor_split`.
    imag : bool, optional
        Enable imaginary time evolution. Defaults to false.

    See Also
    --------
    quimb.Evolution
    """

    def __init__(self, p0, H, dt=None, tol=None, t0=0.0,
                 split_opts=None, progbar=True, imag=False):
        # prepare initial state
        self._pt = p0.copy()
        self._pt.canonize(0)
        self.L = self._pt.L

        # handle hamiltonian -> convert array to LocalHam1D
        if isinstance(H, np.ndarray):
            H = LocalHam1D(L=self.L, H2=H, cyclic=p0.cyclic)

        if not isinstance(H, LocalHam1D):
            raise TypeError("``H`` should be a ``LocalHam1D`` or 2-site "
                            "array, not a TensorNetwork of any form.")

        if p0.cyclic != H.cyclic:
            raise ValueError("Both ``p0`` and ``H`` should have matching OBC "
                             "or PBC.")

        self.H = H
        self.cyclic = H.cyclic
        self._ham_norm = H.mean_norm()
        self._err = 0.0

        # set time and tolerance defaults
        self.t0 = self.t = t0
        if dt and tol:
            raise ValueError("Can't set default for both ``dt`` and ``tol``.")
        self.dt = self._dt = dt
        self.tol = tol
        self.imag = imag

        # misc other options
        self.progbar = progbar
        self.split_opts = ensure_dict(split_opts)

    @property
    def pt(self):
        """The MPS state of the system at the current time.
        """
        return self._pt.copy()

    @property
    def err(self):
        return self._err

    def choose_time_step(self, tol, T, order):
        """Trotter error is ``~ (T / dt) * dt^(order + 1)``. Invert to
        find desired time step, and scale by norm of interaction term.
        """
        return (tol / (T * self._ham_norm)) ** (1 / order)

    def _get_gate_from_ham(self, dt_frac, sites):
        """Get the unitary (exponentiated) gate for fraction of timestep
        ``dt_frac`` and sites ``sites``, cached.
        """
        imag_factor = 1.0 if self.imag else 1.0j
        return self.H.get_gate_expm(sites, -imag_factor * self._dt * dt_frac)

    def sweep(self, direction, dt_frac, dt=None, queue=False):
        """Perform a single sweep of gates and compression. This shifts the
        orthonognality centre along with the gates as they are applied and
        split.

        Parameters
        ----------
        direction : {'right', 'left'}
            Which direction to sweep. Right is even bonds, left is odd.
        dt_frac : float
            What fraction of dt substep to take.
        dt : float, optional
            Overide the current ``dt`` with a custom value.
        """

        # if custom dt set, scale the dt fraction
        if dt is not None:
            dt_frac *= (dt / self._dt)

        # ------ automatically combine consecutive sweeps of same time ------ #

        if not hasattr(self, '_queued_sweep'):
            self._queued_sweep = None

        if queue:
            # check for queued sweep
            if self._queued_sweep:
                # if matches, combine and continue
                if direction == self._queued_sweep[0]:
                    self._queued_sweep[1] += dt_frac
                    return
                # else perform the old, queue the new
                else:
                    new_queued_sweep = [direction, dt_frac]
                    direction, dt_frac = self._queued_sweep
                    self._queued_sweep = new_queued_sweep

            # just queue the new sweep
            else:
                self._queued_sweep = [direction, dt_frac]
                return

        # check if need to drain the queue first
        elif self._queued_sweep:
            queued_direction, queued_dt_frac = self._queued_sweep
            self._queued_sweep = None
            self.sweep(queued_direction, queued_dt_frac, queue=False)

        # ------------------------------------------------------------------- #

        if direction == 'right':
            start_site_ind = 0
            final_site_ind = self.L - 1
            # Apply even gates:
            #
            #     o-<-<-<-<-<-<-<-<-<-   -<-<
            #     | | | | | | | | | |     | |       >~>~>~>~>~>~>~>~>~>~>~o
            #     UUU UUU UUU UUU UUU ... UUU  -->  | | | | | | | | | | | |
            #     | | | | | | | | | |     | |
            #      1   2   3   4   5  ==>
            #
            for i in range(start_site_ind, final_site_ind, 2):
                sites = (i, (i + 1) % self.L)
                U = self._get_gate_from_ham(dt_frac, sites)
                self._pt.left_canonize(start=max(0, i - 1), stop=i)
                self._pt.gate_split_(
                    U, where=sites, absorb='right', **self.split_opts)

            if (self.L % 2 == 1):
                self._pt.left_canonize_site(self.L - 2)
                if self.cyclic:
                    sites = (self.L - 1, 0)
                    U = self._get_gate_from_ham(dt_frac, sites)
                    self._pt.right_canonize_site(1)
                    self._pt.gate_split_(
                        U, where=sites, absorb='left', **self.split_opts)

        elif direction == 'left':

            if self.cyclic and (self.L % 2 == 0):
                sites = (self.L - 1, 0)
                U = self._get_gate_from_ham(dt_frac, sites)
                self._pt.right_canonize_site(1)
                self._pt.gate_split_(
                    U, where=sites, absorb='left', **self.split_opts)

            final_site_ind = 1
            # Apply odd gates:
            #
            #     >->->-   ->->->->->->->->-o
            #     | | |     | | | | | | | | |       o~<~<~<~<~<~<~<~<~<~<~<
            #     | UUU ... UUU UUU UUU UUU |  -->  | | | | | | | | | | | |
            #     | | |     | | | | | | | | |
            #           <==  4   3   2   1
            #
            for i in reversed(range(final_site_ind, self.L - 1, 2)):
                sites = (i, (i + 1) % self.L)
                U = self._get_gate_from_ham(dt_frac, sites)
                self._pt.right_canonize(
                    start=min(self.L - 1, i + 2), stop=i + 1)
                self._pt.gate_split_(
                    U, where=sites, absorb='left', **self.split_opts)

            # one extra canonicalization not included in last split
            self._pt.right_canonize_site(1)

        # Renormalise after imaginary time evolution
        if self.imag:
            factor = self._pt[final_site_ind].norm()
            self._pt[final_site_ind] /= factor

    def _step_order2(self, tau=1, **sweep_opts):
        """Perform a single, second order step.
        """
        self.sweep('right', tau / 2, **sweep_opts)
        self.sweep('left', tau, **sweep_opts)
        self.sweep('right', tau / 2, **sweep_opts)

    def _step_order4(self, **sweep_opts):
        """Perform a single, fourth order step.
        """
        tau1 = tau2 = 1 / (4 * 4**(1 / 3))
        tau3 = 1 - 2 * tau1 - 2 * tau2
        self._step_order2(tau1, **sweep_opts)
        self._step_order2(tau2, **sweep_opts)
        self._step_order2(tau3, **sweep_opts)
        self._step_order2(tau2, **sweep_opts)
        self._step_order2(tau1, **sweep_opts)

    def step(self, order=2, dt=None, progbar=None, **sweep_opts):
        """Perform a single step of time ``self.dt``.
        """
        {2: self._step_order2,
         4: self._step_order4}[order](dt=dt, **sweep_opts)

        dt = self._dt if dt is None else dt
        self.t += dt
        self._err += self._ham_norm * dt ** (order + 1)

        if progbar is not None:
            progbar.cupdate(self.t)
            self._set_progbar_desc(progbar)

    def _compute_sweep_dt_tol(self, T, dt, tol, order):
        # Work out timestep, possibly from target tol, and checking defaults
        dt = self.dt if (dt is None) else dt
        tol = self.tol if (tol is None) else tol

        if not (dt or tol):
            raise ValueError("Must set one of ``dt`` and ``tol``.")
        if (dt and tol):
            raise ValueError("Can't set both ``dt`` and ``tol``.")

        if dt is None:
            self._dt = self.choose_time_step(tol, T - self.t, order)
        else:
            self._dt = dt

        return self._dt

    TARGET_TOL = 1e-13  # tolerance to have 'reached' target time

    def update_to(self, T, dt=None, tol=None, order=4, progbar=None):
        """Update the state to time ``T``.

        Parameters
        ----------
        T : float
            The time to evolve to.
        dt : float, optional
            Time step to use. Can't be set as well as ``tol``.
        tol : float, optional
            Tolerance for whole evolution. Can't be set as well as ``dt``.
        order : int, optional
            Trotter order to use.
        progbar : bool, optional
            Manually turn the progress bar off.
        """
        if T < self.t - self.TARGET_TOL:
            # can't go backwards yet
            raise NotImplementedError

        self._compute_sweep_dt_tol(T, dt, tol, order)

        # set up progress bar and start evolution
        progbar = self.progbar if (progbar is None) else progbar
        progbar = continuous_progbar(self.t, T) if progbar else None

        while self.t < T - self._dt:
            # get closer until we can reach in a single step
            self.step(order=order, progbar=progbar, dt=None, queue=True)

        # always perform final sweep with queue draining
        self.step(order=order, progbar=progbar, dt=T - self.t, queue=False)

        if progbar:
            progbar.close()

    def _set_progbar_desc(self, progbar):
        msg = f"t={self.t:.4g}, max-bond={self._pt.max_bond()}"
        progbar.set_description(msg)

    def at_times(self, ts, dt=None, tol=None, order=4, progbar=None):
        """Generate the time evolved state at each time in ``ts``.

        Parameters
        ----------
        ts : sequence of float
            The times to evolve to and yield the state at.
        dt : float, optional
            Time step to use. Can't be set as well as ``tol``.
        tol : float, optional
            Tolerance for whole evolution. Can't be set as well as ``dt``.
        order : int, optional
            Trotter order to use.
        progbar : bool, optional
            Manually turn the progress bar off.

        Yields
        ------
        pt : MatrixProductState
            The state at each of the times in ``ts``. This is a copy of
            internal state used, so inplace changes can be made to it.
        """
        # convert ts to list, to to calc range and use progress bar
        ts = sorted(ts)
        T = ts[-1]

        # need to use dt always so tol applies over whole T sweep
        dt = self._compute_sweep_dt_tol(T, dt, tol, order)

        # set up progress bar
        progbar = self.progbar if (progbar is None) else progbar
        if progbar:
            ts = Progbar(ts)

        for t in ts:
            self.update_to(t, dt=dt, tol=False, order=order, progbar=False)

            if progbar:
                self._set_progbar_desc(ts)

            yield self.pt


def OTOC_local(psi0, H, H_back, ts, i, A, j=None, B=None,
               initial_eigenstate='check', **tebd_opts):
    """ The out-of-time-ordered correlator (OTOC) generating by two local
    operator A and B acting on site 'i', note it's a function of time.

    Parameters
    ----------
    psi0 : MatrixProductState
        The initial state in MPS form.
    H : LocalHam1D
        The Hamiltonian for forward time-evolution.
    H_back : LocalHam1D
        The Hamiltonian for backward time-evolution, should have only
        sign difference with 'H'.
    ts : sequence of float
        The time to evolve to.
    i : int
        The site where the local operators acting on.
    A : array
        The operator to act with.
    initial_eigenstate: {'check', Flase, True}
        To check the psi0 is or not eigenstate of operator B. If psi0 is the
        eigenstate of B, it will run a simpler version of OTOC calculation
        automatically.

    Returns
    ----------
    The OTOC <A(t)B(0)A(t)B(0)>
    """

    if B is None:
        B = A
    if j is None:
        j = i

    if initial_eigenstate == 'check':
        psi = psi0.gate(B, j, contract=True)
        x = psi0.H.expec(psi)
        y = psi.H.expec(psi)
        if abs(x**2 - y) < 1e-10:
            initial_eigenstate = True
        else:
            initial_eigenstate = False

    if initial_eigenstate is True:
        tebd1 = TEBD(psi0, H, **tebd_opts)
        x = psi0.H.expec(psi0.gate(B, j, contract=True))
        for t in ts:
            # evolve forward
            tebd1.update_to(t)
            # apply first A-gate
            psi_t_A = tebd1.pt.gate(A, i, contract=True)
            # evolve backwards
            tebd2 = TEBD(psi_t_A, H_back, **tebd_opts)
            tebd2.update_to(t)
            # compute expectation with second B-gate
            psi_f = tebd2.pt
            yield x * psi_f.H.expec(psi_f.gate(B, j, contract=True))
    else:
        # set the initial TEBD and apply the first operator A to right
        psi0_L = psi0
        tebd1_L = TEBD(psi0_L, H, **tebd_opts)

        psi0_R = psi0.gate(B, j, contract=True)
        tebd1_R = TEBD(psi0_R, H, **tebd_opts)

        for t in ts:
            # evolve forward
            tebd1_L.update_to(t)
            tebd1_R.update_to(t)

            # apply the opertor A to both left and right states
            psi_t_L_A = tebd1_L.pt.gate(A, i, contract=True)
            psi_t_R_A = tebd1_R.pt.gate(A.H, i, contract=True)

            # set the second left and right TEBD
            tebd2_L = TEBD(psi_t_L_A, H_back, **tebd_opts)
            tebd2_R = TEBD(psi_t_R_A, H_back, **tebd_opts)

            # evolve backwards
            tebd2_L.update_to(t)
            tebd2_R.update_to(t)

            # apply the laste operator B to left and compute overlap
            psi_f_L = tebd2_L.pt.gate(B.H, j, contract=True)
            psi_f_R = tebd2_R.pt
            yield psi_f_L.H.expec(psi_f_R)



from pytest import fixture, mark, raises
import numpy as np
from numpy.testing import assert_allclose
import scipy.sparse as sp

from quimb import (
    qu,
    rand_matrix,
    rand_ket,
)
from quimb.core import (
    qarray,
    ensure_qarray,
    issparse,
    isdense,
    isket,
    isop,
    isbra,
    isvec,
    isherm,
    mul,
    dot,
    vdot,
    rdot,
    ldmul,
    rdmul,
    outer,
    explt,
    make_immutable,
    realify,
    dot_sparse,
    par_dot_csr_matvec,
    kron_dense,
    kron_sparse,
)
from quimb.core import kron, kronpow


# ----------------------------- FIXTURES ------------------------------------ #

_SPARSE_FORMATS = ("csr", "bsr", "csc", "coo")
_TEST_SZ = 4


@fixture
def mat_d():
    return rand_matrix(_TEST_SZ)


@fixture
def mat_d2():
    return rand_matrix(_TEST_SZ)


@fixture
def mat_d3():
    return rand_matrix(_TEST_SZ)


@fixture
def mat_s():
    return rand_matrix(_TEST_SZ, sparse=True, density=0.5)


@fixture
def mat_s2():
    return rand_matrix(_TEST_SZ, sparse=True, density=0.5)


@fixture
def ket_d():
    return rand_ket(_TEST_SZ)


@fixture
def ket_d2():
    return rand_ket(_TEST_SZ)


@fixture
def l1d():
    return np.random.randn(_TEST_SZ) + 1.0j * np.random.randn(_TEST_SZ)


@fixture
def mat_s_nnz():
    return rand_matrix(_TEST_SZ, sparse=True, density=0.75)


# --------------------------------------------------------------------------- #
#                                  TESTS                                      #
# --------------------------------------------------------------------------- #

class TestMakeImmutable():
    def test_dense(self):
        mat = qu([[1, 2], [3, 4]])
        make_immutable(mat)
        with raises(ValueError):
            mat[-1, -1] = 1

    @mark.parametrize("stype", _SPARSE_FORMATS)
    def test_sparse(self, stype):
        mat = qu([[1, 2], [3, 4]], stype=stype)
        make_immutable(mat)
        if stype in {'csr', 'csc'}:
            with raises(ValueError):
                mat[-1, -1] = 1


class TestEnsureQarray:
    def test_ensure_qarray(self):
        def foo(n):
            return np.random.randn(n, n)
        a = foo(2)
        assert not isinstance(a, qarray)

        @ensure_qarray
        def foo2(n):
            return np.random.randn(n, n)
        a = foo2(2)
        assert isinstance(a, qarray)


class TestRealify:
    def test_realify(self):
        def foo(a, b):
            return a + 1j * b
        a = foo(1, 1e-15)
        assert a.real == 1
        assert a.imag == 1e-15

        @realify
        def foo2(a, b):
            return a + 1j * b
        a = foo2(1, 1e-15)
        assert a.real == 1
        assert a.imag == 0

    def test_wrong_type(self):
        @realify
        def foo(a, b):
            return str(a) + str(b)
        assert foo(1, 2) == '12'


class TestShapes:
    def test_sparse(self):
        x = np.array([[1], [0]])
        assert not issparse(x)
        assert isdense(x)
        x = sp.csr_matrix(x)
        assert issparse(x)

    def test_ket(self):
        x = np.array([[1], [0]])
        assert(isket(x))
        assert(not isbra(x))
        assert(not isop(x))
        assert isvec(x)
        x = sp.csr_matrix(x)
        assert(isket(x))
        assert isvec(x)
        assert(not isbra(x))
        assert(not isop(x))

    def test_bra(self):
        x = np.array([[1, 0]])
        assert(not isket(x))
        assert(isbra(x))
        assert(not isop(x))
        assert isvec(x)
        x = sp.csr_matrix(x)
        assert(not isket(x))
        assert(isbra(x))
        assert(not isop(x))
        assert isvec(x)

    def test_op(self):
        x = np.array([[1, 0], [0, 0]])
        assert(not isket(x))
        assert(not isbra(x))
        assert(isop(x))
        assert (not isvec(x))
        x = sp.csr_matrix(x)
        assert(not isket(x))
        assert(not isbra(x))
        assert(isop(x))
        assert (not isvec(x))

    def test_isherm(self):
        a = np.array([[1.0, 2.0 + 3.0j],
                      [2.0 - 3.0j, 1.0]])
        assert(isherm(a))
        a = np.array([[1.0, 2.0 - 3.0j],
                      [2.0 - 3.0j, 1.0]])
        assert(not isherm(a))

    def test_isherm_sparse(self):
        a = sp.csr_matrix([[1.0, 2.0 + 3.0j],
                           [2.0 - 3.0j, 1.0]])
        assert(isherm(a))
        a = sp.csr_matrix([[1.0, 2.0 - 3.0j],
                           [2.0 - 3.0j, 1.0]])
        assert(not isherm(a))


class TestMul:
    def test_mul_dense_same(self, mat_d, mat_d2):
        ca = mul(mat_d, mat_d2)
        assert isinstance(ca, qarray)
        cn = np.multiply(mat_d, mat_d2)
        assert_allclose(ca, cn)

    def test_mul_broadcast(self, mat_d, ket_d):
        ca = mul(mat_d, ket_d)
        assert isinstance(ca, qarray)
        cn = np.multiply(mat_d, ket_d)
        assert_allclose(ca, cn)
        ca = mul(mat_d.H, ket_d)
        assert isinstance(ca, qarray)
        cn = np.multiply(mat_d.H, ket_d)
        assert_allclose(ca, cn)

    def test_mul_sparse(self, mat_s, mat_s2):
        cq = mul(mat_s, mat_s2)
        cn = mat_s.A * mat_s2.A
        assert issparse(cq)
        assert_allclose(cq.A, cn)
        cq = mul(mat_s2.A, mat_s)
        cn = mat_s2.A * mat_s.A
        assert issparse(cq)
        assert_allclose(cq.A, cn)

    def test_mul_sparse_broadcast(self, mat_s, ket_d):
        ca = mul(mat_s, ket_d)
        cn = np.multiply(mat_s.A, ket_d)
        assert_allclose(ca.A, cn)
        ca = mul(mat_s.H, ket_d)
        cn = np.multiply(mat_s.H.A, ket_d)
        assert_allclose(ca.A, cn)


class TestDot:
    def test_dot_matrix(self, mat_d, mat_d2):
        ca = dot(mat_d, mat_d2)
        assert isinstance(ca, qarray)
        cn = mat_d @ mat_d2
        assert_allclose(ca, cn)

    def test_dot_ket(self, mat_d, ket_d):
        ca = dot(mat_d, ket_d)
        assert isinstance(ca, qarray)
        cn = mat_d @ ket_d
        assert_allclose(ca, cn)

    def test_dot_sparse_sparse(self, mat_s, mat_s2):
        cq = dot(mat_s, mat_s2)
        cn = mat_s @ mat_s2
        assert issparse(cq)
        assert_allclose(cq.A, cn.A)

    def test_dot_sparse_dense(self, mat_s, ket_d):
        cq = dot(mat_s, ket_d)
        assert isinstance(cq, qarray)
        cq = mat_s @ ket_d
        assert isinstance(cq, qarray)
        cn = mat_s._mul_vector(ket_d)
        assert not issparse(cq)
        assert isdense(cq)
        assert_allclose(cq.A.ravel(), cn)

    def test_dot_sparse_dense_ket(self, mat_s, ket_d):
        cq = dot(mat_s, ket_d)
        cn = mat_s @ ket_d
        assert not issparse(cq)
        assert isdense(cq)
        assert isket(cq)
        assert_allclose(cq.A, cn)

    def test_par_dot_csr_matvec(self, mat_s, ket_d):
        x = par_dot_csr_matvec(mat_s, ket_d)
        y = dot_sparse(mat_s, ket_d)
        assert x.dtype == complex
        assert x.shape == (_TEST_SZ, 1)
        assert isinstance(x, qarray)
        assert_allclose(x, y)

    def test_par_dot_csr_matvec_Array(self, mat_s, ket_d):
        x = par_dot_csr_matvec(mat_s, np.asarray(ket_d).reshape(-1))
        y = dot_sparse(mat_s, ket_d)
        assert x.dtype == complex
        assert x.shape == (_TEST_SZ,)
        assert_allclose(y, x.reshape(-1, 1))


class TestAccelVdot:
    def test_accel_vdot(self, ket_d, ket_d2):
        ca = vdot(ket_d, ket_d2)
        cn = (ket_d.H @ ket_d2)[0, 0]
        assert_allclose(ca, cn)


class TestAccelRdot:
    def test_accel_rdot(self, ket_d, ket_d2):
        cq = rdot(ket_d.H, ket_d2)
        cn = (ket_d.H @ ket_d2)[0, 0]
        assert_allclose(cq, cn)


class TestFastDiagMul:
    def test_ldmul_small(self, mat_d, l1d):
        a = ldmul(l1d, mat_d)
        b = np.diag(l1d) @ mat_d
        assert isinstance(a, qarray)
        assert_allclose(a, b)

    def test_ldmul_large(self):
        vec = np.random.randn(501)
        mat = rand_matrix(501)
        a = ldmul(vec, mat)
        b = np.diag(vec) @ mat
        assert isinstance(a, qarray)
        assert_allclose(a, b)

    def test_ldmul_sparse(self, mat_s, l1d):
        assert issparse(mat_s)
        a = ldmul(l1d, mat_s)
        b = np.diag(l1d) @ mat_s.A
        assert issparse(a)
        assert_allclose(a.A, b)

    def test_rdmul_small(self, mat_d, l1d):
        a = rdmul(mat_d, l1d)
        b = mat_d @ np.diag(l1d)
        assert isinstance(a, qarray)
        assert_allclose(a, b)

    def test_rdmul_large(self):
        vec = np.random.randn(501)
        mat = rand_matrix(501)
        a = rdmul(mat, vec)
        b = mat @ np.diag(vec)
        assert isinstance(a, qarray)
        assert_allclose(a, b)

    def test_rdmul_sparse(self, mat_s, l1d):
        a = rdmul(mat_s, l1d)
        b = mat_s.A @ np.diag(l1d)
        assert issparse(a)
        assert_allclose(a.A, b)


class TestOuter:
    def test_outer_ket_ket(self, ket_d, ket_d2):
        c = outer(ket_d, ket_d2)
        assert isinstance(c, qarray)
        d = np.multiply(ket_d, ket_d2.T)
        assert_allclose(c, d)

    def test_outer_ket_bra(self, ket_d, ket_d2):
        c = outer(ket_d, ket_d2.H)
        assert isinstance(c, qarray)
        d = np.multiply(ket_d, ket_d2.H)
        assert_allclose(c, d)

    def test_outer_bra_ket(self, ket_d, ket_d2):
        c = outer(ket_d.H, ket_d2)
        assert isinstance(c, qarray)
        d = np.multiply(ket_d.H.T, ket_d2.T)
        assert_allclose(c, d)

    def test_outer_bra_bra(self, ket_d, ket_d2):
        c = outer(ket_d.H, ket_d2.H)
        assert isinstance(c, qarray)
        d = np.multiply(ket_d.H.T, ket_d2.H)
        assert_allclose(c, d)


class TestExplt:
    def test_small(self):
        evals = np.random.randn(3)
        en = np.exp(-1.0j * evals * 7)
        eq = explt(evals, 7)
        assert_allclose(eq, en)


# --------------------------------------------------------------------------- #
# Kronecker (tensor) product tests                                            #
# --------------------------------------------------------------------------- #

class TestKron:
    @mark.parametrize("big", [False, True])
    def test_kron_dense(self, mat_d, mat_d2, big):
        x = kron_dense(mat_d, mat_d2, par_thresh=0 if big else 1e100)
        assert mat_d.shape == (_TEST_SZ, _TEST_SZ)
        assert mat_d2.shape == (_TEST_SZ, _TEST_SZ)
        xn = np.kron(mat_d, mat_d2)
        assert_allclose(x, xn)
        assert isinstance(x, qarray)

    def test_kron_multi_args(self, mat_d, mat_d2, mat_d3):
        assert_allclose(kron(mat_d), mat_d)
        assert_allclose(kron(mat_d, mat_d2, mat_d3),
                        np.kron(np.kron(mat_d, mat_d2), mat_d3))

    def test_kron_mixed_types(self, mat_d, mat_s):
        assert_allclose(kron(mat_d, mat_s).A,
                        (sp.kron(mat_d, mat_s, 'csr')).A)
        assert_allclose(kron(mat_s, mat_s).A,
                        (sp.kron(mat_s, mat_s, 'csr')).A)


class TestKronSparseFormats:
    def test_sparse_sparse_auto(self, mat_s):
        c = kron_sparse(mat_s, mat_s)
        assert c.format == 'csr'

    def test_sparse_dense_auto(self, mat_s, mat_d):
        c = kron_sparse(mat_s, mat_d)
        assert c.format == 'bsr'

    def test_dense_sparse_auto(self, mat_s, mat_d):
        c = kron_sparse(mat_d, mat_s)
        assert c.format == 'csr'

    def test_sparse_sparsennz(self, mat_s, mat_s_nnz):
        c = kron_sparse(mat_s, mat_s_nnz)
        assert c.format == 'csr'

    @mark.parametrize("stype", _SPARSE_FORMATS)
    def test_sparse_sparse_to_sformat(self, mat_s, stype):
        c = kron_sparse(mat_s, mat_s, stype=stype)
        assert c.format == stype

    @mark.parametrize("stype", (None,) + _SPARSE_FORMATS)
    def test_many_args_dense_last(self, mat_s, mat_s2, mat_d, stype):
        c = kron(mat_s, mat_s2, mat_d, stype=stype)
        assert c.format == (stype if stype is not None else "bsr")

    @mark.parametrize("stype", (None,) + _SPARSE_FORMATS)
    def test_many_args_dense_not_last(self, mat_s, mat_s2, mat_d, stype):
        c = kron(mat_d, mat_s, mat_s2, stype=stype)
        assert c.format == (stype if stype is not None else "csr")
        c = kron(mat_s, mat_d, mat_s2, stype=stype)
        assert c.format == (stype if stype is not None else "csr")

    @mark.parametrize("stype", (None,) + _SPARSE_FORMATS)
    def test_many_args_dense_last_coo_construct(self, mat_s, mat_s2, mat_d,
                                                stype):
        c = kron(mat_s, mat_s2, mat_d, stype=stype, coo_build=True)
        assert c.format == (stype if stype is not None else "csr")

    @mark.parametrize("stype", (None,) + _SPARSE_FORMATS)
    def test_many_args_dense_not_last_coo_construct(self, mat_s, mat_s2, mat_d,
                                                    stype):
        c = kron(mat_s, mat_d, mat_s2, stype=stype, coo_build=True)
        assert c.format == (stype if stype is not None else "csr")
        c = kron(mat_d, mat_s, mat_s2, stype=stype, coo_build=True)
        assert c.format == (stype if stype is not None else "csr")


class TestKronPow:
    def test_dense(self, mat_d):
        x = mat_d & mat_d & mat_d
        y = kronpow(mat_d, 3)
        assert_allclose(x, y)

    def test_sparse(self, mat_s):
        x = mat_s & mat_s & mat_s
        y = kronpow(mat_s, 3)
        assert_allclose(x.A, y.A)

    @mark.parametrize("stype", _SPARSE_FORMATS)
    def test_sparse_formats(self, stype, mat_s):
        x = mat_s & mat_s & mat_s
        y = kronpow(mat_s, 3, stype=stype)
        assert y.format == stype
        assert_allclose(x.A, y.A)

    @mark.parametrize("sformat_in", _SPARSE_FORMATS)
    @mark.parametrize("stype", (None,) + _SPARSE_FORMATS)
    def test_sparse_formats_coo_construct(self, sformat_in, stype, mat_s):
        mat_s = mat_s.asformat(sformat_in)
        x = mat_s & mat_s & mat_s
        y = kronpow(mat_s, 3, stype=stype, coo_build=True)
        assert y.format == stype if stype is not None else "sformat_in"
        assert_allclose(x.A, y.A)



import random
import itertools
import collections

from autoray import do, to_numpy, dag

from ..core import eye, kron, qarray
from ..utils import ensure_dict
from ..utils import progbar as Progbar
from .tensor_core import Tensor
from .drawing import get_colors, get_positions


class LocalHamGen:
    """Representation of a local hamiltonian defined on a general graph. This
    combines all two site and one site terms into a single interaction per
    lattice pair, and caches operations on the terms such as getting their
    exponential. The sites (nodes) should be hashable and comparable.

    Parameters
    ----------
    H2 : dict[tuple[node], array_like]
        The interaction terms, with each key being an tuple of nodes defining
        an edge and each value the local hamilotonian term for those two nodes.
    H1 : array_like or dict[node, array_like], optional
        The one site term(s). If a single array is given, assume to be the
        default onsite term for all terms. If a dict is supplied,
        the keys should represent specific coordinates like
        ``(i, j)`` with the values the array representing the local term for
        that site. A default term for all remaining sites can still be supplied
        with the key ``None``.

    Attributes
    ----------
    terms : dict[tuple, array_like]
        The total effective local term for each interaction (with single site
        terms appropriately absorbed). Each key is a pair of coordinates
        ``site_a, site_b`` with ``site_a < site_b``.
    """

    def __init__(self, H2, H1=None):
        # caches for not repeating operations / duplicating tensors
        self._op_cache = collections.defaultdict(dict)

        self.terms = dict(H2)

        # convert qarrays (mostly useful for working with jax)
        for key, X in self.terms.items():
            if isinstance(X, qarray):
                self.terms[key] = self._convert_from_qarray_cached(X)

        self.sites = tuple(
            sorted(set(itertools.chain.from_iterable(self.terms)))
        )

        # first combine terms to ensure coo1 < coo2
        for where in tuple(filter(bool, self.terms)):
            coo1, coo2 = where
            if coo1 < coo2:
                continue

            # pop and flip the term
            X12 = self._flip_cached(self.terms.pop(where))

            # add to, or create, term with flipped coos
            new_where = coo2, coo1
            if new_where in self.terms:
                self.terms[new_where] = self._add_cached(
                    self.terms[new_where], X12
                )
            else:
                self.terms[new_where] = X12

        # make a directory of which single sites are covered by which terms
        #     - to merge them into later
        self._sites_to_covering_terms = collections.defaultdict(list)
        for where in self.terms:
            site_a, site_b = where
            self._sites_to_covering_terms[site_a].append(where)
            self._sites_to_covering_terms[site_b].append(where)

        # parse one site terms
        if H1 is None:
            H1s = dict()
        elif hasattr(H1, "shape"):
            # set a default site term
            H1s = {None: H1}
        else:
            H1s = dict(H1)

        # convert qarrays (mostly useful for working with jax)
        for key, X in H1s.items():
            if isinstance(X, qarray):
                H1s[key] = self._convert_from_qarray_cached(X)

        # possibly set the default single site term
        default_H1 = H1s.pop(None, None)
        if default_H1 is not None:
            for site in self.sites:
                H1s.setdefault(site, default_H1)

        # now absorb the single site terms evenly into the two site terms
        for site, H in H1s.items():

            # get interacting terms which cover the site
            pairs = self._sites_to_covering_terms[site]
            num_pairs = len(pairs)
            if num_pairs == 0:
                raise ValueError(
                    f"There are no two site terms to add this single site "
                    f"term to - site {site} is not coupled to anything."
                )

            # merge the single site term in equal parts into all covering pairs
            H_tensoreds = (self._op_id_cached(H), self._id_op_cached(H))
            for pair in pairs:
                H_tensored = H_tensoreds[pair.index(site)]
                self.terms[pair] = self._add_cached(
                    self.terms[pair], self._div_cached(H_tensored, num_pairs)
                )

    @property
    def nsites(self):
        """The number of sites in the system.
        """
        return len(self.sites)

    def items(self):
        """Iterate over all terms in the hamiltonian. This is mostly for
        convenient compatibility with ``compute_local_expectation``.
        """
        return self.terms.items()

    def _convert_from_qarray_cached(self, x):
        cache = self._op_cache["convert_from_qarray"]
        key = id(x)
        if key not in cache:
            cache[key] = x.A
        return cache[key]

    def _flip_cached(self, x):
        cache = self._op_cache["flip"]
        key = id(x)
        if key not in cache:
            d = int(x.size ** (1 / 4))
            xf = do("reshape", x, (d, d, d, d))
            xf = do("transpose", xf, (1, 0, 3, 2))
            xf = do("reshape", xf, (d * d, d * d))
            cache[key] = xf
        return cache[key]

    def _add_cached(self, x, y):
        cache = self._op_cache["add"]
        key = (id(x), id(y))
        if key not in cache:
            cache[key] = x + y
        return cache[key]

    def _div_cached(self, x, y):
        cache = self._op_cache["div"]
        key = (id(x), y)
        if key not in cache:
            cache[key] = x / y
        return cache[key]

    def _op_id_cached(self, x):
        cache = self._op_cache["op_id"]
        key = id(x)
        if key not in cache:
            xn = to_numpy(x)
            d = int(xn.size ** 0.5)
            Id = eye(d, dtype=xn.dtype)
            XI = do("array", kron(xn, Id), like=x)
            cache[key] = XI
        return cache[key]

    def _id_op_cached(self, x):
        cache = self._op_cache["id_op"]
        key = id(x)
        if key not in cache:
            xn = to_numpy(x)
            d = int(xn.size ** 0.5)
            Id = eye(d, dtype=xn.dtype)
            IX = do("array", kron(Id, xn), like=x)
            cache[key] = IX
        return cache[key]

    def _expm_cached(self, x, y):
        cache = self._op_cache["expm"]
        key = (id(x), y)
        if key not in cache:
            el, ev = do("linalg.eigh", x)
            cache[key] = ev @ do("diag", do("exp", el * y)) @ dag(ev)
        return cache[key]

    def get_gate(self, where):
        """Get the local term for pair ``where``, cached.
        """
        return self.terms[tuple(sorted(where))]

    def get_gate_expm(self, where, x):
        """Get the local term for pair ``where``, matrix exponentiated by
        ``x``, and cached.
        """
        return self._expm_cached(self.get_gate(where), x)

    def apply_to_arrays(self, fn):
        """Apply the function ``fn`` to all the arrays representing terms.
        """
        for k, x in self.terms.items():
            self.terms[k] = fn(x)

    def _nx_color_ordering(self, strategy="smallest_first", interchange=True):
        """Generate a term ordering based on a coloring on the line graph.
        """
        import networkx as nx

        G = nx.Graph(tuple(self.terms))

        coloring = list(
            nx.coloring.greedy_color(
                nx.line_graph(G), strategy, interchange=interchange
            ).items()
        )

        # sort into color groups
        coloring.sort(key=lambda coo_color: coo_color[1])

        return [
            # networkx doesn't preserve node order of edge spec
            tuple(sorted(coo)) for
            coo, _ in coloring
        ]

    def get_auto_ordering(self, order="sort", **kwargs):
        """Get an ordering of the terms to use with TEBD, for example. The
        default is to sort the coordinates then greedily group them into
        commuting sets.

        Parameters
        ----------
        order : {'sort', None, 'random', str}
            How to order the terms *before* greedily grouping them into
            commuting (non-coordinate overlapping) sets:

                - ``'sort'`` will sort the coordinate pairs first.
                - ``None`` will use the current order of terms which should
                  match the order they were supplied to this ``LocalHam2D``
                  instance.
                - ``'random'`` will randomly shuffle the coordinate pairs
                  before grouping them - *not* the same as returning a
                  completely random order.
                - ``'random-ungrouped'`` will randomly shuffle the coordinate
                  pairs but *not* group them at all with respect to
                  commutation.

            Any other option will be passed as a strategy to
            ``networkx.coloring.greedy_color`` to generate the ordering.

        Returns
        -------
        list[tuple[node]]
            Sequence of coordinate pairs.
        """
        if order is None:
            pairs = self.terms
        elif order == "sort":
            pairs = sorted(self.terms)
        elif order == "random":
            pairs = list(self.terms)
            random.shuffle(pairs)
        elif order == "random-ungrouped":
            pairs = list(self.terms)
            random.shuffle(pairs)
            return pairs
        else:
            return self._nx_color_ordering(order, **kwargs)

        pairs = {x: None for x in pairs}

        cover = set()
        ordering = list()
        while pairs:
            for pair in tuple(pairs):
                ij1, ij2 = pair
                if (ij1 not in cover) and (ij2 not in cover):
                    ordering.append(pair)
                    pairs.pop(pair)
                    cover.add(ij1)
                    cover.add(ij2)
            cover.clear()

        return ordering

    def __repr__(self):
        s = "<LocalHamGen(nsites={}, num_terms={})>"
        return s.format(self.nsites, len(self.terms))

    def draw(
        self,
        ordering="sort",
        show_norm=True,
        figsize=None,
        fontsize=8,
        legend=True,
        ax=None,
        return_fig=False,
        **kwargs,
    ):
        """Plot this Hamiltonian as a network.

        Parameters
        ----------
        ordering : {'sort', None, 'random'}, optional
            An ordering of the termns, or an argument to be supplied to
            :meth:`quimb.tensor.tensor_arbgeom_tebd.LocalHamGen.get_auto_ordering`
            to generate this automatically.
        show_norm : bool, optional
            Show the norm of each term as edge labels.
        figsize : None or tuple[int], optional
            Size of the figure, defaults to size of Hamiltonian.
        fontsize : int, optional
            Font size for norm labels.
        legend : bool, optional
            Whether to show the legend of which terms are in which group.
        ax : None or matplotlib.Axes, optional
            Add to a existing set of axes.
        return_fig : bool, optional
            Whether to return any newly created figure.
        """
        import networkx as nx
        import matplotlib.pyplot as plt

        if figsize is None:
            L = self.nsites ** 0.5 + 1
            figsize = (L, L)

        ax_supplied = ax is not None
        if not ax_supplied:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            ax.axis("off")
            ax.set_aspect("equal")

        if ordering is None or isinstance(ordering, str):
            ordering = self.get_auto_ordering(ordering, **kwargs)

        G = nx.Graph()
        seen = set()
        n = 0
        edge_labels = dict()
        for where in ordering:
            site_a, site_b = where
            if (site_a in seen) or (site_b in seen):
                # start a new group
                seen = {site_a, site_b}
                n += 1
            else:
                seen.add(site_a)
                seen.add(site_b)

            nrm = do("linalg.norm", self.terms[where])
            edge_labels[where] = f"{nrm:.2f}"
            G.add_edge(site_a, site_b, norm=nrm, group=n)

        num_groups = n + 1
        colors = get_colors(range(num_groups))

        pos = get_positions(None, G)

        # do the plotting
        nx.draw_networkx_edges(
            G,
            pos=pos,
            width=tuple(2 * x[2]["norm"] ** 0.5 for x in G.edges(data=True)),
            edge_color=tuple(
                colors[x[2]["group"]] for x in G.edges(data=True)
            ),
            alpha=0.8,
            ax=ax,
        )
        nx.draw_networkx_edge_labels(
            G,
            pos=pos,
            edge_labels=edge_labels,
            font_size=fontsize,
            font_color=(0.5, 0.5, 0.5),
            bbox=dict(alpha=0),
            ax=ax,
        )
        nx.draw_networkx_labels(
            G,
            pos,
            font_color=(0.2, 0.2, 0.2),
            font_size=fontsize,
            font_weight="bold",
            ax=ax,
        )

        # create legend
        if legend:
            handles = []
            for color in colors.values():
                handles += [
                    plt.Line2D(
                        [0],
                        [0],
                        marker="o",
                        color=color,
                        linestyle="",
                        markersize=10,
                    )
                ]

            lbls = [f"Group {i + 1}" for i in range(num_groups)]

            ax.legend(
                handles,
                lbls,
                ncol=max(round(len(handles) / 20), 1),
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )

        if ax_supplied:
            return

        if return_fig:
            return fig

        plt.show()

    graph = draw


class TEBDGen:
    """Generic class for performing time evolving block decimation on an
    arbitrary graph, i.e. applying the exponential of a Hamiltonian using
    a product formula that involves applying local exponentiated gates only.
    """

    def __init__(
        self,
        psi0,
        ham,
        tau=0.01,
        D=None,
        imag=True,
        gate_opts=None,
        ordering=None,
        second_order_reflect=False,
        compute_energy_every=None,
        compute_energy_final=True,
        compute_energy_opts=None,
        compute_energy_fn=None,
        compute_energy_per_site=False,
        callback=None,
        keep_best=False,
        progbar=True,
    ):
        self.imag = imag
        if not imag:
            raise NotImplementedError("Real time evolution not tested yet.")

        self.state = psi0
        self.ham = ham
        self.progbar = progbar
        self.callback = callback

        # default time step to use
        self.tau = tau

        # parse gate application options
        if D is None:
            D = self._psi.max_bond()
        self.gate_opts = ensure_dict(gate_opts)
        self.gate_opts['max_bond'] = D
        self.gate_opts.setdefault('cutoff', 0.0)
        self.gate_opts.setdefault('contract', 'reduce-split')

        # parse energy computation options
        self.compute_energy_opts = ensure_dict(compute_energy_opts)

        self.compute_energy_every = compute_energy_every
        self.compute_energy_final = compute_energy_final
        self.compute_energy_fn = compute_energy_fn
        self.compute_energy_per_site = bool(compute_energy_per_site)

        if ordering is None:

            def dynamic_random():
                return self.ham.get_auto_ordering('random_sequential')

            self.ordering = dynamic_random
        elif isinstance(ordering, str):
            self.ordering = self.ham.get_auto_ordering(ordering)
        elif callable(ordering):
            self.ordering = ordering
        else:
            self.ordering = tuple(ordering)

        self.second_order_reflect = second_order_reflect

        # storage
        self._n = 0
        self.its = []
        self.taus = []
        self.energies = []

        self.keep_best = bool(keep_best)
        self.best = dict(energy=float('inf'), state=None, it=None)

    def sweep(self, tau):
        r"""Perform a full sweep of gates at every pair.

        .. math::

            \psi \rightarrow \prod_{\{ij\}} \exp(-\tau H_{ij}) \psi

        """
        if callable(self.ordering):
            ordering = self.ordering()
        else:
            ordering = self.ordering

        if self.second_order_reflect:
            ordering = tuple(ordering) + tuple(reversed(ordering))
            factor = 2.0
        else:
            factor = 1.0

        for where in ordering:

            if callable(tau):
                U = self.ham.get_gate_expm(where, -tau(where) / factor)
            else:
                U = self.ham.get_gate_expm(where, -tau / factor)

            self.gate(U, where)

    def _update_progbar(self, pbar):
        desc = f"n={self._n}, tau={self.tau}, energy~{float(self.energy):.6f}"
        pbar.set_description(desc)

    def evolve(self, steps, tau=None, progbar=None):
        """Evolve the state with the local Hamiltonian for ``steps`` steps with
        time step ``tau``.
        """
        if tau is not None:
            self.tau = tau

        if progbar is None:
            progbar = self.progbar

        pbar = Progbar(total=steps, disable=self.progbar is not True)

        try:
            for i in range(steps):
                # anything required by both energy and sweep
                self.presweep(i)

                # possibly compute the energy
                should_compute_energy = (
                    bool(self.compute_energy_every) and
                    (i % self.compute_energy_every == 0))
                if should_compute_energy:
                    self._check_energy()
                    self._update_progbar(pbar)

                # actually perform the gates
                self.sweep(self.tau)
                self._n += 1
                pbar.update()

                if self.callback is not None:
                    if self.callback(self):
                        break

            # possibly compute the energy
            if self.compute_energy_final:
                self._check_energy()
                self._update_progbar(pbar)

        except KeyboardInterrupt:
            # allow the user to interupt early
            pass
        finally:
            pbar.close()

    @property
    def state(self):
        """Return a copy of the current state.
        """
        return self.get_state()

    @state.setter
    def state(self, psi):
        self.set_state(psi)

    @property
    def n(self):
        """The number of sweeps performed.
        """
        return self._n

    @property
    def D(self):
        """The maximum bond dimension.
        """
        return self.gate_opts['max_bond']

    @D.setter
    def D(self, value):
        """The maximum bond dimension.
        """
        self.gate_opts['max_bond'] = round(value)

    def _check_energy(self):
        """Logic for maybe computing the energy if needed.
        """
        if self.its and (self._n == self.its[-1]):
            # only compute if haven't already
            return self.energies[-1]

        if self.compute_energy_fn is not None:
            en = self.compute_energy_fn(self)
        else:
            en = self.compute_energy()

        if self.compute_energy_per_site:
            en = en / self.ham.nsites

        self.energies.append(float(en))
        self.taus.append(float(self.tau))
        self.its.append(self._n)

        if self.keep_best and en < self.best['energy']:
            self.best['energy'] = en
            self.best['state'] = self.state
            self.best['it'] = self._n

        return self.energies[-1]

    @property
    def energy(self):
        """Return the energy of current state, computing it only if necessary.
        """
        return self._check_energy()

    # ------- abstract methods that subclasses might want to override ------- #

    def get_state(self):
        """The default method for retrieving the current state - simply a copy.
        Subclasses can override this to perform additional transformations.
        """
        return self._psi.copy()

    def set_state(self, psi):
        """The default method for setting the current state - simply a copy.
        Subclasses can override this to perform additional transformations.
        """
        self._psi = psi.copy()

    def presweep(self, i):
        """Perform any computations required before the sweep (and energy
        computation). For the basic TEBD this is nothing.
        """
        pass

    def gate(self, U, where):
        """Perform single gate ``U`` at coordinate pair ``where``. This is the
        the most common method to override.
        """
        self._psi.gate_(U, where, **self.gate_opts)

    def compute_energy(self):
        """Compute and return the energy of the current state. Subclasses can
        override this with a custom method to compute the energy.
        """
        return self._psi.compute_local_expectation_simple(
            terms=self.ham.terms,
            **self.compute_energy_opts
        )

    def __repr__(self):
        s = "<{}(n={}, tau={}, D={})>"
        return s.format(
            self.__class__.__name__, self.n, self.tau, self.D)


class SimpleUpdateGen(TEBDGen):

    def gate(self, U, where):
        self._psi.gate_simple_(
            U, where, gauges=self.gauges, **self.gate_opts
        )

    def compute_energy(self):
        return self._psi.compute_local_expectation_simple(
            terms=self.ham.terms,
            gauges=self.gauges,
            **self.compute_energy_opts,
        )

    def get_state(self, absorb_gauges=True):
        psi = self._psi.copy()

        if absorb_gauges:
            psi.gauge_simple_insert(self.gauges)
        else:
            for ix, g in self.gauges.items():
                psi |= Tensor(g, inds=[ix])

        return psi

    def set_state(self, psi):
        self._psi = psi.copy()
        self.gauges = {}
        self._psi.gauge_all_simple_(gauges=self.gauges)



import pytest
import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn


def rand_rect(m, n, sparse=False, dtype=complex):
    X = qu.rand_matrix(max(m, n), dtype=dtype, sparse=sparse)
    return X[:m, :n]


def usv2dense(U, s, VH):
    return U @ np.diag(s) @ VH


def rand_rank(m, n, k, dtype=complex):
    s = np.sort(qu.randn(k)**2)[::-1]

    U = qu.gen.rand.rand_iso(m, k, dtype=dtype)
    VH = qu.gen.rand.rand_iso(n, k, dtype=dtype).conj().T

    if U.dtype in ('float32', 'complex64'):
        s = s.astype('float32')

    return usv2dense(U, s, VH)


def rand_tn1d_sect(n, bd, dtype=complex):
    mps = qtn.MPS_rand_state(n + 2, bd, dtype=dtype)
    mpo = qtn.MPO_rand_herm(n + 2, 5, dtype=dtype)

    norm = qtn.TensorNetwork(qtn.tensor_network_align(mps.H, mpo, mps))
    norm.view_as_(qtn.TensorNetwork1D, like=mps)

    lix = qtn.bonds(norm[0], norm[1])
    rix = qtn.bonds(norm[n], norm[n + 1])

    to = norm[1:n + 1]

    return qtn.TNLinearOperator1D(to, lix, rix, 1, n + 1)


dtypes = ['float32', 'float64', 'complex64', 'complex128']


class TestRSVD:

    @pytest.mark.parametrize('dtype', dtypes)
    @pytest.mark.parametrize('shape', [(41, 31), (31, 41)])
    @pytest.mark.parametrize('sparse', [False, True])
    @pytest.mark.parametrize('q', [2, 3])
    @pytest.mark.parametrize('p', [0, 5])
    def test_rsvd(self, dtype, shape, sparse, q, p):
        X = rand_rect(*shape, dtype=dtype, sparse=sparse)

        k = 15
        U, s, V = qu.rsvd(X, k, q=q, p=p)

        assert U.shape == (shape[0], k)
        assert s.shape == (k,)
        assert V.shape == (k, shape[1])

        assert U.dtype == dtype
        assert V.dtype == dtype

        assert_allclose(U.conj().T @ U, np.eye(k), rtol=1e-5, atol=1e-5)
        assert_allclose(V @ V.conj().T, np.eye(k), rtol=1e-5, atol=1e-5)

        Ue, se, Ve = qu.svds(X, k)
        opt_err = qu.norm(X.A - usv2dense(Ue, se, Ve), 'fro')
        act_err = qu.norm(X.A - usv2dense(U, s, V), 'fro')

        assert act_err < 1.2 * opt_err

        assert_allclose(s[:k // 2], se[:k // 2], rtol=0.05)

    @pytest.mark.parametrize('dtype', dtypes)
    @pytest.mark.parametrize('shape', [(41, 31), (31, 41)])
    @pytest.mark.parametrize('q', [2, 3])
    @pytest.mark.parametrize('p', [0, 5])
    def test_rsvd_adaptive(self, dtype, shape, q, p):
        X = rand_rank(*shape, 10, dtype=dtype)
        U, s, V = qu.rsvd(X, 1e-6, q=q, p=p, k_start=10)

        k = s.size
        assert 10 <= k <= 20

        assert U.dtype == dtype
        assert V.dtype == dtype

        assert_allclose(U.conj().T @ U, np.eye(k), rtol=1e-6, atol=1e-6)
        assert_allclose(V @ V.conj().T, np.eye(k), rtol=1e-6, atol=1e-6)

        Ue, se, Ve = qu.svds(X, k)
        act_err = qu.norm(X - usv2dense(U, s, V), 'fro')

        assert act_err < 1e-4

        assert_allclose(s[:k // 2], se[:k // 2], rtol=0.1)

    @pytest.mark.parametrize('dtype', dtypes)
    @pytest.mark.parametrize('shape', [(410, 310), (310, 410)])
    @pytest.mark.parametrize('k_start', [4, 10, 16])
    @pytest.mark.parametrize('use_qb', [False, 10, True])
    def test_estimate_rank(self, dtype, shape, k_start, use_qb):
        rnk = 100
        X = rand_rank(*shape, rnk, dtype=dtype)

        Ue, se, VHe = qu.svd(X)
        assert_allclose(se[rnk:], 0.0, atol=1e-5)

        k = qu.estimate_rank(X, 1e-3, k_start=k_start,
                             use_qb=use_qb, use_sli=False)
        assert_allclose(k, 100, rtol=0.3)

        assert qu.estimate_rank(X, 1e-3, k_start=k_start, k_max=50) == 50

    @pytest.mark.parametrize('dtype', dtypes)
    @pytest.mark.parametrize('k_start', [2, 4, 8])
    @pytest.mark.parametrize('use_qb', [False, 10, True])
    def test_estimate_rank_lo(self, dtype, k_start, use_qb):
        X = rand_tn1d_sect(30, 10, dtype=dtype)

        Ue, se, VHe = qu.svd(X.A)
        actual_rank = sum(se > se[0] * 1e-3)

        k = qu.estimate_rank(X, 1e-3, k_start=k_start,
                             use_qb=use_qb, use_sli=False)
        assert_allclose(k, actual_rank, rtol=0.3 if use_qb else 0.5)

        assert qu.estimate_rank(X, 1e-3, k_start=k_start, k_max=8) == 8



import pytest
import numpy as np
from numpy.testing import assert_allclose
import quimb as qu


@pytest.mark.parametrize("sparse", [False, True])
@pytest.mark.parametrize("stype", ['csr', 'csc'])
@pytest.mark.parametrize("dtype", [
    "don't pass", None, np.float64, np.complex128])
def test_hamiltonian_builder(sparse, stype, dtype):
    from quimb.gen.operators import hamiltonian_builder

    @hamiltonian_builder
    def simple_ham(sparse=None, stype=None, dtype=None):
        H = qu.qu([[0.0, 1.0],
                   [1.0, 0.0]], sparse=True, stype='csr', dtype=dtype)
        return H

    @hamiltonian_builder
    def simple_ham_complex(sparse=None, stype=None, dtype=None):
        H = qu.qu([[0.0, 1.0j],
                   [-1.0j, 0.0]], sparse=True, stype='csr', dtype=dtype)
        return H

    if dtype == "don't pass":
        H = simple_ham(sparse=sparse, stype=stype)
    else:
        H = simple_ham(sparse=sparse, stype=stype, dtype=dtype)

    if dtype == "don't pass" or dtype is None:
        # check that passng no actual dtype keeps it as float
        assert H.dtype == np.float64
    else:
        # check that explicit dtypes are respected
        assert H.dtype == dtype
    assert qu.issparse(H) == sparse
    assert qu.isdense(H) != sparse
    if sparse:
        assert H.format == stype

    with pytest.raises(ValueError):  # check immutability
        H[0, 0] = 100

    if dtype == "don't pass":
        H = simple_ham_complex(sparse=sparse, stype=stype)
    elif dtype is np.float64:
        with pytest.warns(np.ComplexWarning):
            H = simple_ham_complex(sparse=sparse, stype=stype, dtype=dtype)
    else:
        H = simple_ham_complex(sparse=sparse, stype=stype, dtype=dtype)
    if dtype == "don't pass" or dtype is None:
        # check that passng no actual dtype keeps it as float
        assert H.dtype == np.complex128
    else:
        # check that explicit dtypes are respected
        assert H.dtype == dtype
    assert qu.issparse(H) == sparse
    assert qu.isdense(H) != sparse
    if sparse:
        assert H.format == stype

    with pytest.raises(ValueError):  # check immutability
        H[0, 0] = 100

    return


class TestSpinOperator:
    def test_spin_half(self):
        Sx = qu.spin_operator('x', 1 / 2)
        assert_allclose(Sx, [[0.0, 0.5], [0.5, 0.0]])

        Sy = qu.spin_operator('y', 1 / 2)
        assert_allclose(Sy, [[0.0, -0.5j], [0.5j, 0.0]])

        Sz = qu.spin_operator('z', 1 / 2)
        assert_allclose(Sz, [[0.5, 0.0], [0.0, -0.5]])

        Sp = qu.spin_operator('+', 1 / 2)
        assert_allclose(Sp, [[0.0, 1.0], [0.0, 0.0]])

        Sm = qu.spin_operator('-', 1 / 2)
        assert_allclose(Sm, [[0.0, 0.0], [1.0, 0.0]])

        SI = qu.spin_operator('I', 1 / 2)
        assert_allclose(SI, [[1.0, 0.0], [0.0, 1.0]])

    @pytest.mark.parametrize("label", ('x', 'y', 'z'))
    @pytest.mark.parametrize("S", [1, 3 / 2, 2, 5 / 2])
    def test_spin_high(self, label, S):
        D = int(2 * S + 1)
        op = qu.spin_operator(label, S)
        assert_allclose(qu.eigvalsh(op), np.linspace(-S, S, D), atol=1e-13)


class TestPauli:
    def test_pauli_dim2(self):
        for dir in (1, 'x', 'X',
                    2, 'y', 'Y',
                    3, 'z', 'Z'):
            x = qu.pauli(dir)
            assert_allclose(qu.eigvalsh(x), [-1, 1])

    def test_pauli_dim3(self):
        for dir in (1, 'x', 'X',
                    2, 'y', 'Y',
                    3, 'z', 'Z'):
            x = qu.pauli(dir, dim=3)
            assert_allclose(qu.eigvalsh(x), [-1, 0, 1],
                            atol=1e-15)

    def test_pauli_bad_dim(self):
        with pytest.raises(KeyError):
            qu.pauli('x', 4)

    def test_pauli_bad_dir(self):
        with pytest.raises(KeyError):
            qu.pauli('w', 2)


class TestControlledZ:
    def test_controlled_z_dense(self):
        cz = qu.controlled('z')
        assert_allclose(cz, np.diag([1, 1, 1, -1]))

    def test_controlled_z_sparse(self):
        cz = qu.controlled('z', sparse=True)
        assert(qu.issparse(cz))
        assert_allclose(cz.A, np.diag([1, 1, 1, -1]))


class TestGates:

    @pytest.mark.parametrize("gate", ['Rx', 'Ry', 'Rz', 'T_gate', 'S_gate',
                                      'CNOT', 'cX', 'cY', 'cZ', 'hadamard',
                                      'phase_gate', 'iswap', 'swap', 'U_gate',
                                      'fsim', 'fsimg'])
    @pytest.mark.parametrize('dtype', [np.complex64, np.complex128])
    @pytest.mark.parametrize('sparse', [False, True])
    def test_construct(self, gate, dtype, sparse):
        if gate in {'Rx', 'Ry', 'Rz', 'phase_gate'}:
            args = (0.43827,)
        elif gate in {'U_gate'}:
            args = (0.1, 0.2, 0.3)
        elif gate in {'fsim'}:
            args = (-1.3, 5.4)
        elif gate in {'fsimg'}:
            args = (-1.3, 5.4, 2., 3., 4.)
        else:
            args = ()
        G = getattr(qu, gate)(*args, dtype=dtype, sparse=sparse)
        assert G.dtype == dtype
        assert qu.issparse(G) is sparse
        psi = qu.rand_ket(G.shape[0])
        Gpsi = G @ psi
        assert qu.expec(Gpsi, Gpsi) == pytest.approx(1.0)

    def test_gates_import(self):
        from quimb.gates import Z
        assert_allclose(Z, [[1, 0], [0, -1]])

    def test_fsim(self):
        assert_allclose(qu.fsim(- qu.pi / 2, 0.0), qu.iswap(), atol=1e-12)

    def test_fsimg(self):
        assert_allclose(
            qu.fsimg(- qu.pi / 2, 0.0, 0.0, 0.0, 0.0),
            qu.iswap(), atol=1e-12
        )


class TestHamHeis:
    def test_ham_heis_2(self):
        h = qu.ham_heis(2, cyclic=False)
        evals = qu.eigvalsh(h)
        assert_allclose(evals, [-0.75, 0.25, 0.25, 0.25])
        gs = qu.groundstate(h)
        assert_allclose(qu.expec(gs, qu.singlet()), 1.)

    @pytest.mark.parametrize("parallel", [False, True])
    def test_ham_heis_sparse_cyclic_4(self, parallel):
        h = qu.ham_heis(4, sparse=True, cyclic=True, parallel=parallel)
        lk = qu.eigvalsh(h, k=4)
        assert_allclose(lk, [-2, -1, -1, -1])

    def test_ham_heis_bz(self):
        h = qu.ham_heis(2, cyclic=False, b=1)
        evals = qu.eigvalsh(h)
        assert_allclose(evals, [-3 / 4, -3 / 4, 1 / 4, 5 / 4])

    @pytest.mark.parametrize("stype", ["coo", "csr", "csc", "bsr"])
    def test_sformat_construct(self, stype):
        h = qu.ham_heis(4, sparse=True, stype=stype)
        assert h.format == stype


class TestHamJ1J2:
    def test_ham_j1j2_3_dense(self):
        h = qu.ham_j1j2(3, j2=1.0, cyclic=False)
        h2 = qu.ham_heis(3, cyclic=True)
        assert_allclose(h, h2)

    def test_ham_j1j2_6_sparse_cyc(self):
        h = qu.ham_j1j2(6, j2=0.5, sparse=True, cyclic=True)
        lk = qu.eigvalsh(h, k=5)
        assert_allclose(lk, [-9 / 4, -9 / 4, -7 / 4, -7 / 4, -7 / 4])

    def test_ham_j1j2_4_bz(self):
        h = qu.ham_j1j2(4, j2=0.5, cyclic=True, bz=0)
        lk = qu.eigvalsh(h, k=11)
        assert_allclose(lk, [-1.5, -1.5, -0.5, -0.5, -0.5, -0.5,
                             -0.5, -0.5, -0.5, -0.5, -0.5])
        h = qu.ham_j1j2(4, j2=0.5, cyclic=True, bz=0.05)
        lk = qu.eigvalsh(h, k=11)
        assert_allclose(lk, [-1.5, -1.5, -0.55, -0.55, -0.55,
                             -0.5, -0.5, -0.5, -0.45, -0.45, -0.45])


class TestHamMBL:
    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("sparse", [False, True])
    @pytest.mark.parametrize("dh_dim", [1, 2, 3, 'y', 'xz'])
    @pytest.mark.parametrize("dh_dist", ['s', 'g'])
    def test_construct(self, cyclic, sparse, dh_dim, dh_dist):
        qu.ham_mbl(n=3, dh=3, cyclic=cyclic, sparse=sparse, dh_dim=dh_dim,
                   dh_dist=dh_dist)

    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("sparse", [False, True])
    def test_construct_qp(self, cyclic, sparse):
        qu.ham_mbl(n=3, dh=3, cyclic=cyclic, sparse=sparse, dh_dist='qp')


class TestHamHeis2D:
    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("sparse", [False, True])
    @pytest.mark.parametrize("parallel", [False, True])
    @pytest.mark.parametrize("bz", [0.0, 0.7])
    def test_construct(self, cyclic, sparse, parallel, bz):
        qu.ham_heis_2D(2, 3, cyclic=cyclic, sparse=sparse,
                       parallel=parallel, bz=bz)


class TestSpinZProjector:
    @pytest.mark.parametrize("sz", [-2, -1, 0, 1, 2])
    def test_works(self, sz):
        prj = qu.zspin_projector(4, sz)
        h = qu.ham_heis(4)
        h0 = prj.T @ h @ prj
        v0s = qu.eigvecsh(h0)
        for i in range(v0s.shape[1]):
            v0 = v0s[:, [i]]
            vf = prj @ v0
            prjv = vf @ vf.H
            # Check reconstructed full eigenvectors commute with full ham
            assert_allclose(prjv @ h, h @ prjv, atol=1e-13)
        if sz == 0:
            # Groundstate must be in most symmetric subspace
            gs = qu.groundstate(h)
            gs0 = prj @ v0s[:, [0]]
            assert_allclose(qu.expec(gs, gs0), 1.0)
            assert_allclose(qu.expec(h, gs0), qu.expec(h, gs))

    def test_raises(self):
        with pytest.raises(ValueError):
            qu.zspin_projector(5, 0)
        with pytest.raises(ValueError):
            qu.zspin_projector(4, 1 / 2)

    @pytest.mark.parametrize("sz", [(-1 / 2, 1 / 2), (3 / 2, 5 / 2)])
    def test_spin_half_double_space(self, sz):
        prj = qu.zspin_projector(5, sz)
        h = qu.ham_heis(5)
        h0 = prj.T @ h @ prj
        v0s = qu.eigvecsh(h0)
        for i in range(v0s.shape[1]):
            v0 = v0s[:, [i]]
            vf = prj @ v0
            prjv = vf @ vf.H
            # Check reconstructed full eigenvectors commute with full ham
            assert_allclose(prjv @ h, h @ prjv, atol=1e-13)
        if sz == 0:
            # Groundstate must be in most symmetric subspace
            gs = qu.groundstate(h)
            gs0 = prj @ v0s[:, [0]]
            assert_allclose(qu.expec(gs, gs0), 1.0)
            assert_allclose(qu.expec(h, gs0), qu.expec(h, gs))


class TestSwap:
    @pytest.mark.parametrize("sparse", [False, True])
    def test_swap_qubits(self, sparse):
        a = qu.up() & qu.down()
        s = qu.swap(2, sparse=sparse)
        assert_allclose(s @ a, qu.down() & qu.up())

    @pytest.mark.parametrize("sparse", [False, True])
    def test_swap_higher_dim(self, sparse):
        a = qu.rand_ket(9)
        s = qu.swap(3, sparse=sparse)
        assert_allclose(s @ a, a.reshape([3, 3]).T.reshape([9, 1]))


@pytest.mark.parametrize("sparse", [False, True])
def test_3qubit_gates(sparse):
    psi = qu.rand_ket(8)
    psi = qu.toffoli(sparse=sparse) @ psi
    psi = qu.ccY(sparse=sparse) @ psi
    psi = qu.fredkin(sparse=sparse) @ psi
    psi = qu.ccZ(sparse=sparse) @ psi
    assert qu.expec(psi, psi) == pytest.approx(1.0)


class TestHubbardSpinless:

    def test_half_filling_groundstate(self):
        H = qu.ham_hubbard_hardcore(8, t=0.5, V=1.0, mu=1.0, cyclic=True)
        gs = qu.groundstate(H)
        dims = [2] * 8
        cn = qu.num(2)
        ens = [qu.expec(cn, qu.ptr(gs, dims, i)) for i in range(8)]
        for en in ens:
            assert en == pytest.approx(0.5, rel=1e-6)



from pytest import fixture, mark, raises

from math import pi, gcd, cos
from functools import reduce

import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
from quimb.evo import (
    schrodinger_eq_ket,
    schrodinger_eq_dop,
    schrodinger_eq_dop_vectorized,
    lindblad_eq,
    lindblad_eq_vectorized,
)
from .test_linalg.test_slepc_linalg import slepc4py_test

from quimb.linalg.base_linalg import eigs_scipy


@fixture
def psi_dot():
    psi = qu.rand_ket(3)
    ham = 10 * qu.rand_herm(3)
    psid = -1.0j * (ham @ psi)
    return psi, ham, psid


@fixture
def spsi_dot():
    psi = qu.rand_ket(3)
    ham = qu.rand_herm(3, sparse=True, density=0.5)
    psid = -1.0j * (ham @ psi)
    return psi, ham, psid


@fixture
def rho_dot():
    rho = qu.rand_rho(3)
    ham = qu.rand_herm(3)
    rhod = -1.0j * (ham @ rho - rho @ ham)
    return rho, ham, rhod


@fixture
def srho_dot():
    rho = qu.rand_rho(3)
    ham = qu.rand_herm(3, sparse=True, density=0.5)
    rhod = -1.0j * (ham @ rho - rho @ ham)
    return rho, ham, rhod


@fixture
def rho_dot_ls():
    np.random.seed(1)
    rho = qu.rand_rho(3)
    ham = qu.rand_herm(3)
    gamma = 0.7
    ls = [qu.rand_matrix(3) for _ in range(3)]
    rhodl = -1.0j * (ham @ rho - rho @ ham)
    for l in ls:
        rhodl += gamma * (l @ rho @ l.H)
        rhodl -= gamma * 0.5 * (rho @ l.H @ l)
        rhodl -= gamma * 0.5 * (l.H @ l @ rho)
    return rho, ham, gamma, ls, rhodl


@fixture
def srho_dot_ls():
    rho = qu.rand_rho(3)
    ham = qu.rand_herm(3, sparse=True, density=0.5)
    gamma = 0.7
    ls = [qu.rand_matrix(3, sparse=True, density=0.5) for _ in range(3)]
    rhodl = -1.0j * (ham @ rho - rho @ ham)
    for l in ls:
        rhodl += gamma * (l @ rho @ l.H)
        rhodl -= gamma * 0.5 * (rho @ l.H @ l)
        rhodl -= gamma * 0.5 * (l.H @ l @ rho)
    return rho, ham, gamma, ls, rhodl


# --------------------------------------------------------------------------- #
# Evolution equation tests                                                    #
# --------------------------------------------------------------------------- #

class TestSchrodingerEqKet:
    def test_ket_matrix(self, psi_dot):
        psi, ham, psid = psi_dot
        foo = schrodinger_eq_ket(ham)
        psid2 = foo(None, psi)
        assert_allclose(psid, psid2)

    def test_ket_1darray(self, psi_dot):
        psi, ham, psid = psi_dot
        foo = schrodinger_eq_ket(ham)
        psid2 = foo(None, psi.A.reshape(-1)).reshape(-1, 1)
        assert_allclose(psid, psid2)

    def test_ket_matrix_sparse(self, spsi_dot):
        psi, ham, psid = spsi_dot
        foo = schrodinger_eq_ket(ham)
        psid2 = foo(None, psi)
        assert_allclose(psid, psid2)

    def test_ket_1darray_sparse(self, spsi_dot):
        psi, ham, psid = spsi_dot
        foo = schrodinger_eq_ket(ham)
        psid2 = foo(None, psi.A.reshape(-1)).reshape(-1, 1)
        assert_allclose(psid, psid2)


class TestSchrodingerEqDop:
    def test_dop_matrix(self, rho_dot):
        rho, ham, rhod = rho_dot
        foo = schrodinger_eq_dop(ham)
        rhod2 = foo(None, rho.A).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_dop_1darray(self, rho_dot):
        rho, ham, rhod = rho_dot
        foo = schrodinger_eq_dop(ham)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_dop_matrix_sparse(self, srho_dot):
        rho, ham, rhod = srho_dot
        foo = schrodinger_eq_dop(ham)
        rhod2 = foo(None, rho.A).reshape(3, 3)
        assert_allclose(rhod, rhod2, atol=1e-12)

    def test_dop_1darray_sparse(self, srho_dot):
        rho, ham, rhod = srho_dot
        foo = schrodinger_eq_dop(ham)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2, atol=1e-12)


class TestSchrodingerEqDopVec:
    def test_dop_1darray(self, rho_dot):
        rho, ham, rhod = rho_dot
        foo = schrodinger_eq_dop_vectorized(ham)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_dop_1darray_sparse(self, srho_dot):
        rho, ham, rhod = srho_dot
        foo = schrodinger_eq_dop_vectorized(ham)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2, atol=1e-12)


class TestLindbladEq:
    def test_matrix(self, rho_dot_ls):
        rho, ham, gamma, ls, rhod = rho_dot_ls
        foo = lindblad_eq(ham, ls, gamma)
        rhod2 = foo(None, rho).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_1darray(self, rho_dot_ls):
        rho, ham, gamma, ls, rhod = rho_dot_ls
        foo = lindblad_eq(ham, ls, gamma)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_matrix_sparse(self, srho_dot_ls):
        rho, ham, gamma, ls, rhod = srho_dot_ls
        foo = lindblad_eq(ham, ls, gamma)
        rhod2 = foo(None, rho).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_1darray_sparse(self, srho_dot_ls):
        rho, ham, gamma, ls, rhod = srho_dot_ls
        foo = lindblad_eq(ham, ls, gamma)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)


class TestLindbladEqVec:
    def test_1darray(self, rho_dot_ls):
        rho, ham, gamma, ls, rhod = rho_dot_ls
        foo = lindblad_eq_vectorized(ham, ls, gamma)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)

    def test_1darray_sparse(self, srho_dot_ls):
        rho, ham, gamma, ls, rhod = srho_dot_ls
        foo = lindblad_eq_vectorized(ham, ls, gamma)
        rhod2 = foo(None, rho.A.reshape(-1)).reshape(3, 3)
        assert_allclose(rhod, rhod2)


# --------------------------------------------------------------------------- #
# Evolution class tests                                                       #
# --------------------------------------------------------------------------- #


@fixture
def ham_rcr_psi():
    # Define a random hamiltonian with a known recurrence time
    d = 3
    np.random.seed(1)
    ems = np.random.randint(1, 6, d)
    ens = np.random.randint(1, 6, d)  # eigenvalues as rational numbers
    # numerator lowest common divisor
    LCD = reduce(gcd, ems)
    # denominator lowest common multiple
    LCM = reduce(lambda a, b: a * b // gcd(a, b), ens)
    trc = 2 * pi * LCM / LCD
    evals = np.array(ems) / np.array(ens)
    v = qu.rand_uni(d)
    ham = v @ np.diag(evals) @ v.H
    p0 = qu.rand_ket(d)
    tm = 0.573 * trc
    pm = v @ np.diag(np.exp(-1.0j * tm * evals)) @ v.H @ p0
    return ham, trc, p0, tm, pm


class TestEvolution:
    @mark.parametrize("sparse, presolve",
                      [(False, False),
                       (True, False),
                       (False, True)])
    def test_evo_ham_dense_ket_solve(self, ham_rcr_psi, sparse, presolve):
        ham, trc, p0, tm, pm = ham_rcr_psi
        ham = qu.qu(ham, sparse=sparse)
        if presolve:
            l, v = qu.eigh(ham)
            sim = qu.Evolution(p0, (l, v))
            assert isinstance(sim._ham, tuple) and len(sim._ham) == 2
        else:
            sim = qu.Evolution(p0, ham, method='solve')
        sim.update_to(tm)
        assert_allclose(sim.pt, pm)
        assert qu.expec(sim.pt, p0) < 1.0
        sim.update_to(trc)
        assert_allclose(sim.pt, p0)
        assert isinstance(sim.pt, qu.qarray)
        assert sim.t == trc

    @mark.parametrize("dop", [False, True])
    @mark.parametrize("sparse", [False, True])
    @mark.parametrize("method", ["solve", "integrate", 'expm', 'bad'])
    @mark.parametrize("timedep", [False, True])
    @mark.parametrize("linop", [False, True])
    def test_evo_ham(self, ham_rcr_psi, sparse, dop, method, timedep, linop):
        ham, trc, p0, tm, pm = ham_rcr_psi
        if dop:
            if method == 'expm':
                # XXX: not implemented
                return
            p0 = p0 @ p0.H
            pm = pm @ pm.H

        if method == 'bad':
            with raises(ValueError):
                qu.Evolution(p0, ham, method=method)
            return

        ham = qu.qu(ham, sparse=sparse)

        if linop:
            import scipy.sparse.linalg as spla

            ham = spla.aslinearoperator(ham)

        if timedep:
            # fake a time dependent ham by making it callable
            ham_object, ham = ham, (lambda t: ham_object)

        if linop and (method in ('expm', 'solve')):
            with raises(TypeError):
                qu.Evolution(p0, ham, method=method)
            return

        if timedep and (method in ('expm', 'solve')):
            with raises(TypeError):
                qu.Evolution(p0, ham, method=method)
            return

        sim = qu.Evolution(p0, ham, method=method)
        sim.update_to(tm)
        assert_allclose(sim.pt, pm, rtol=1e-4, atol=1e-6)
        assert qu.expec(sim.pt, p0) < 1.0
        sim.update_to(trc)
        assert_allclose(sim.pt, p0, rtol=1e-4, atol=1e-6)
        assert isinstance(sim.pt, qu.qarray)
        assert sim.t == trc

    @mark.parametrize("dop", [False, True])
    @mark.parametrize("linop", [False, True])
    @mark.parametrize("num_callbacks", [0, 1, 2])
    @mark.parametrize("use_int_stop", [False, True])
    def test_evo_timedep_adiabatic_with_callbacks(self, dop, linop,
                                                  num_callbacks, use_int_stop):
        # tests time dependent Evolution via an adiabatic sweep with:
        #   a) no callbacks
        #   b) 1 callback that accesses the time-dependent Hamiltonian
        #   c) 2 callbacks where one access the Hamiltonian and one doesn't

        if num_callbacks > 0 and (dop or linop):
            # should implement this at some point
            return

        L = 6
        T = 20

        H1 = qu.ham_mbl(L, dh=1.0, seed=4, sparse=True, cyclic=True)
        gs1 = qu.groundstate(H1)
        H2 = qu.ham_mbl(L, dh=1.0, seed=5, sparse=True, cyclic=True)
        gs2 = qu.groundstate(H2)

        if linop:
            import scipy.sparse.linalg as spla

            H1 = spla.aslinearoperator(H1)
            H2 = spla.aslinearoperator(H2)

        # make sure two ground states are different
        assert qu.fidelity(gs1, gs2) < 0.5

        # linearly interpolate from one ham to the other
        def ham(t):
            return (1 - t / T) * H1 + (t / T) * H2

        if linop:
            assert isinstance(ham(0.3), spla.LinearOperator)

        if dop:
            p0 = qu.dop(gs1)
        else:
            p0 = gs1

        if use_int_stop:
            def check_init_gs_overlap(t, pt):
                val = qu.fidelity(pt, gs1)
                return (-1 if val <= 0.75 else 0)
            int_stop = check_init_gs_overlap
        else:
            int_stop = None

        if num_callbacks == 0:
            evo = qu.Evolution(p0, ham, method='integrate', int_stop=int_stop,
                               progbar=True)
        else:
            def gs_overlap(t, pt, H):
                evals, evecs = eigs_scipy(H(t), k=1, which='SA')
                return np.abs(qu.dot(pt.T, qu.qu(evecs[:, 0])))**2

            if num_callbacks == 1:
                compute = gs_overlap
            if num_callbacks == 2:
                def norm(t, pt):
                    return qu.dot(pt.T, pt)
                compute = {'norm': norm, 'gs_overlap': gs_overlap}
            evo = qu.Evolution(p0, ham, compute=compute, int_stop=int_stop,
                               method='integrate', progbar=True)
        evo.update_to(T)

        # final state should now overlap much more with second hamiltonian GS
        if use_int_stop:
            assert qu.fidelity(evo.pt, gs1) < 0.9
            assert qu.fidelity(evo.pt, gs2) > 0.1
            assert evo.t < 15
        else:
            assert qu.fidelity(evo.pt, gs1) < 0.5
            assert qu.fidelity(evo.pt, gs2) > 0.99
            assert evo.t == 20

        if num_callbacks == 1:
            gs_overlap_results = evo.results
            # check that we stayed in the ground state the whole time
            assert ((np.array(gs_overlap_results) - 1.0) < 1e-3).all()

        if num_callbacks == 2:
            norm_results = evo.results['norm']
            gs_overlap_results = evo.results['gs_overlap']
            # check that we stayed normalized the whole time
            assert ((np.array(norm_results) - 1.0) < 1e-3).all()
            # check that we stayed in the ground state the whole time
            assert ((np.array(gs_overlap_results) - 1.0) < 1e-3).all()

    def test_int_stop_calling_details(self, ham_rcr_psi):
        # test some details about the way Evolution is called with int_stop:
        # - Giving int_stop without any compute
        # - Giving int_stop with (t, p) and with (t, p, H) call signatures
        ham, trc, p0, tm, pm = ham_rcr_psi

        # check that the int_stop argument doesn't get accepted in either form
        with raises(ValueError):
            qu.Evolution(p0, ham, method='solve', int_stop=(lambda t, p: -1))
        with raises(ValueError):
            qu.Evolution(p0, ham, method='solve',
                         int_stop=(lambda t, p, H: -1))

        # check expected behaviour in case where int_stop takes t, p
        sim = qu.Evolution(p0, ham, method='integrate',
                           int_stop=(lambda t, p: -1))
        sim.update_to(trc)
        assert sim.t < trc / 2  # make sure it stopped early

        sim = qu.Evolution(p0, ham, method='integrate',
                           int_stop=(lambda t, p: 0))
        sim.update_to(trc)
        assert sim.t == trc  # make sure it didn't stop early

        sim = qu.Evolution(p0, ham, method='integrate',
                           int_stop=(lambda t, p, H: -1))

        # check expected behaviour in case where int_stop takes t, p, H
        sim.update_to(trc)
        assert sim.t < trc / 2  # make sure it stopped early

        sim = qu.Evolution(p0, ham, method='integrate',
                           int_stop=(lambda t, p, H: 0))
        sim.update_to(trc)
        assert sim.t == trc  # make sure it didn't stop early

        # check that TypeError not related to argument count gets properly
        #   raised
        with raises(TypeError):
            sim = qu.Evolution(p0, ham, method='integrate',
                               int_stop=7)
            sim.update_to(trc)

    def test_evo_at_times(self):
        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.up() & qu.down()
        sim = qu.Evolution(p0, ham, method='solve')
        ts = np.linspace(0, 10)
        for t, pt in zip(ts, sim.at_times(ts)):
            x = cos(t)
            y = qu.expec(pt, qu.ikron(qu.pauli('z'), [2, 2], 0))
            assert_allclose(x, y, atol=1e-15)

    @mark.parametrize("qtype", ['ket', 'dop'])
    @mark.parametrize("method", ['solve', 'integrate', 'expm'])
    def test_evo_compute_callback(self, qtype, method):
        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.qu(qu.up() & qu.down(), qtype=qtype)

        def some_quantity(t, pt):
            return t, qu.logneg(pt)

        evo = qu.Evolution(p0, ham, method=method, compute=some_quantity)
        manual_lns = []
        for pt in evo.at_times(np.linspace(0, 1, 6)):
            manual_lns.append(qu.logneg(pt))
        ts, lns = zip(*evo.results)
        assert len(lns) >= len(manual_lns)
        # check a specific value of logneg at t=0.8 was computed automatically
        checked = False
        for t, ln in zip(ts, lns):
            if abs(t - 0.8) < 1e-12:
                assert abs(ln - manual_lns[4]) < 1e-12
                checked = True
        assert checked

    @mark.parametrize("qtype", ['ket', 'dop'])
    @mark.parametrize("method", ['solve', 'integrate', 'expm'])
    def test_evo_multi_compute(self, method, qtype):

        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.qu(qu.up() & qu.down(), qtype=qtype)

        def some_quantity(t, _):
            return t

        def some_other_quantity(_, pt):
            return qu.logneg(pt)

        # check that hamiltonian gets accepted without error for all methods
        def some_other_quantity_accepting_ham(t, pt, H):
            return qu.logneg(pt)

        compute = {'t': some_quantity, 'logneg': some_other_quantity,
                   'logneg_ham': some_other_quantity_accepting_ham}

        evo = qu.Evolution(p0, ham, method=method, compute=compute)
        manual_lns = []
        for pt in evo.at_times(np.linspace(0, 1, 6)):
            manual_lns.append(qu.logneg(pt))
        ts = evo.results['t']
        lns = evo.results['logneg']
        lns_ham = evo.results['logneg_ham']
        assert len(lns) >= len(manual_lns)
        # check a specific value of logneg at t=0.8 was computed automatically
        checked = False
        for t, ln, ln_ham in zip(ts, lns, lns_ham):
            if abs(t - 0.8) < 1e-12:
                assert abs(ln - manual_lns[4]) < 1e-12
                # check that accepting hamiltonian didn't mess it up
                assert ln == ln_ham
                checked = True
        assert checked

    @slepc4py_test
    @mark.parametrize('expm_backend', ['slepc-krylov', 'slepc-expokit'])
    def test_expm_slepc(self, expm_backend):
        ham = qu.ham_mbl(7, dh=0.5, sparse=True)
        psi = qu.rand_ket(2**7)
        evo_exact = qu.Evolution(psi, ham, method='solve')
        evo_slepc = qu.Evolution(psi, ham, method='expm',
                                 expm_backend=expm_backend)
        ts = np.linspace(0, 100, 6)
        for p1, p2 in zip(evo_exact.at_times(ts), evo_slepc.at_times(ts)):
            assert abs(qu.expec(p1, p2) - 1) < 1e-9

    def test_progbar_update_to_integrate(self, capsys):
        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.up() & qu.down()
        sim = qu.Evolution(p0, ham, method='integrate', progbar=True)
        sim.update_to(100)
        # check something as been printed
        _, err = capsys.readouterr()
        assert err and "%" in err

    def test_progbar_at_times_solve(self, capsys):
        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.up() & qu.down()
        sim = qu.Evolution(p0, ham, method='solve', progbar=True)
        for _ in sim.at_times(np.linspace(0, 100, 11)):
            pass
        # check something as been printed
        _, err = capsys.readouterr()
        assert err and "%" in err

    def test_progbar_at_times_expm(self, capsys):
        ham = qu.ham_heis(2, cyclic=False)
        p0 = qu.up() & qu.down()
        sim = qu.Evolution(p0, ham, method='expm', progbar=True)
        for _ in sim.at_times(np.linspace(0, 100, 11)):
            pass
        # check something as been printed
        _, err = capsys.readouterr()
        assert err and "%" in err
